#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko PDBQT writer
#

import sys
import json
import math
from os import linesep as eol

import numpy as np
from rdkit import Chem
from .utils import pdbutils
from .utils.rdkitutils import mini_periodic_table

from .molsetup import Bond


def oids_json_from_setup(molsetup, name="LigandFromMeeko"):
    if len(molsetup.restraints):
        raise NotImplementedError(
            "molsetup has restraints but these aren't written to oids block yet"
        )
    offchrg_type = "OFFCHRG"
    offchrg_by_parent = {}
    for atom in molsetup.atoms:
        if atom.is_pseudo_atom and atom.atom_type == offchrg_type:
            neigh = molsetup.get_neighbors(atom.index)
            if len(neigh) != 1:
                raise RuntimeError(
                    "offsite charge %s is bonded to: %s which has len() != 1"
                    % (atom.index, json.dumps(neigh))
                )
            if neigh[0] in offchrg_by_parent:
                raise RuntimeError(
                    "atom %d has more than one offsite charge" % neigh[0]
                )
            offchrg_by_parent[neigh[0]] = atom.index
    output_indices_start_at_one = True
    index_start = int(output_indices_start_at_one)
    positions_block = ""
    charges = []
    offchrg_by_oid_parent = {}
    elements = []
    n_real_atoms = molsetup.true_atom_count
    n_fake_atoms = molsetup.pseudoatom_count
    indexmap = {}  # molsetup: oid
    count_oids = 0
    for atom in molsetup.atoms:
        index = atom.index
        if atom.is_dummy or atom.is_pseudo_atom or atom.is_ignore:
            continue
        if atom.atom_type == offchrg_type:
            continue # handled by offchrg_by_parent
        oid_id = count_oids + index_start
        indexmap[index] = count_oids
        x, y, z = atom.coord
        positions_block += "position.%d = (%f,%f,%f)\n" % (oid_id, x, y, z)
        charges.append(atom.charge)
        if index in offchrg_by_parent:
            index_pseudo = offchrg_by_parent[index]
            xq_abs, yq_abs, zq_abs = molsetup.atoms[index_pseudo].coord
            xq_rel = xq_abs - x
            yq_rel = yq_abs - y
            zq_rel = zq_abs - z
            offchrg_by_oid_parent[count_oids] = {
                "q": molsetup.atoms[index_pseudo].charge,
                "xyz": (xq_rel, yq_rel, zq_rel),
            }
        count_oids += 1
        element = "%s %s %d" % (name, atom.atom_type, oid_id)
        elements.append(element)

    tmp = []
    for index in range(len(charges)):
        if index in offchrg_by_oid_parent:
            tmplist = [
                "%f" % charges[index],
                "0.0",
                "0.0",
                "0.0",
            ]  # xyz relative to current elemtn
            tmplist.append("%f" % offchrg_by_oid_parent[index]["q"])
            tmplist.append("%f,%f,%f" % offchrg_by_oid_parent[index]["xyz"])
            tmp.append(",".join(tmplist))
        else:
            tmp.append("%f" % charges[index])
    charges_line = "import_charges = {%s}\n" % ("|".join(tmp))
    elements_line = "elements = %s\n" % (",".join(elements))

    bonds = [[] for _ in range(count_oids)]
    # bond_orders = [[] for _ in range(count_oids)]
    static_links = []
    for i, j in molsetup.bond_info.keys():
        if molsetup.get_is_ignore(i) or molsetup.get_is_ignore(j):
            continue
        if (
            molsetup.get_atom_type(i) == offchrg_type
            or molsetup.get_atom_type(j) == offchrg_type
        ):
            continue
        oid_i = indexmap[i]
        oid_j = indexmap[j]
        bonds[oid_i].append("%d" % (oid_j + index_start))
        # bond_orders[oid_i].append("%d" % molsetup.bond_info[(i, j)].order)
        if not molsetup.bond_info[(i, j)].rotatable:
            static_links.append("%d,%d" % (oid_i + index_start, oid_j + index_start))
    bonds = [",".join(j_list) for j_list in bonds]
    bonds_line = "connectivity = {%s}\n" % ("|".join(bonds))
    # bond_orders = [",".join(orders) for orders in bond_orders]
    # bondorder_line = "bond_order = {%s}\n" % ("|".join(bond_orders))
    staticlinks_line = "static_links = {%s}\n" % ("|".join(static_links))

    output = ""
    output += "[Group: %s]\n" % name
    output += positions_block
    output += charges_line
    output += elements_line
    output += bonds_line
    #output += bondorder_line
    output += staticlinks_line
    output += "number = 1\t\t// can only be 1 for the sandbox currently (but any number for classical MC)\n"
    output += "group_dipole = 1\t// not relevant for sandbox but classical MC\n"
    output += "rand_independent=0\t// not relevant for sandbox but classical MC\n"
    output += "bond_range = 4\t\t// bond range AD default\n"
    output += "\n"
    output += get_dihedrals_block(molsetup, indexmap, name)

    return output, indexmap


def oids_block_from_setup(molsetup, name="LigandFromMeeko"):
    if len(molsetup.restraints):
        raise NotImplementedError(
            "molsetup has restraints but these aren't written to oids block yet"
        )
    offchrg_type = "OFFCHRG"
    offchrg_by_parent = {}
    for atom in molsetup.atoms:
        if atom.is_pseudo_atom and atom.atom_type == offchrg_type:
            neigh = molsetup.get_neighbors(atom.index)
            if len(neigh) != 1:
                raise RuntimeError(
                    "offsite charge %s is bonded to: %s which has len() != 1"
                    % (atom.index, json.dumps(neigh))
                )
            if neigh[0] in offchrg_by_parent:
                raise RuntimeError(
                    "atom %d has more than one offsite charge" % neigh[0]
                )
            offchrg_by_parent[neigh[0]] = atom.index
    output_indices_start_at_one = True
    index_start = int(output_indices_start_at_one)
    positions_block = ""
    charges = []
    offchrg_by_oid_parent = {}
    elements = []
    n_real_atoms = molsetup.true_atom_count
    n_fake_atoms = molsetup.pseudoatom_count
    indexmap = {}  # molsetup: oid
    count_oids = 0
    for atom in molsetup.atoms:
        index = atom.index
        if atom.is_dummy or atom.is_pseudo_atom or atom.is_ignore:
            continue
        if atom.atom_type == offchrg_type:
            continue  # handled by offchrg_by_parent
        oid_id = count_oids + index_start
        indexmap[index] = count_oids
        x, y, z = atom.coord
        positions_block += "position.%d = (%f,%f,%f)\n" % (oid_id, x, y, z)
        charges.append(atom.charge)
        if index in offchrg_by_parent:
            index_pseudo = offchrg_by_parent[index]
            xq_abs, yq_abs, zq_abs = molsetup.atoms[index_pseudo].coord
            xq_rel = xq_abs - x
            yq_rel = yq_abs - y
            zq_rel = zq_abs - z
            offchrg_by_oid_parent[count_oids] = {
                "q": molsetup.get_charge(index_pseudo),
                "xyz": (xq_rel, yq_rel, zq_rel),
            }
        count_oids += 1
        element = "%s %s %d" % (name, molsetup.get_atom_type(index), oid_id)
        elements.append(element)

    tmp = []
    for index in range(len(charges)):
        if index in offchrg_by_oid_parent:
            tmplist = [
                "%f" % charges[index],
                "0.0",
                "0.0",
                "0.0",
            ]  # xyz relative to current elemtn
            tmplist.append("%f" % offchrg_by_oid_parent[index]["q"])
            tmplist.append("%f,%f,%f" % offchrg_by_oid_parent[index]["xyz"])
            tmp.append(",".join(tmplist))
        else:
            tmp.append("%f" % charges[index])
    charges_line = "import_charges = {%s}\n" % ("|".join(tmp))
    elements_line = "elements = %s\n" % (",".join(elements))

    bonds = [[] for _ in range(count_oids)]
    # bond_orders = [[] for _ in range(count_oids)]
    static_links = []
    for i, j in molsetup.bond_info.keys():
        if molsetup.get_is_ignore(i) or molsetup.get_is_ignore(j):
            continue
        if (
            molsetup.get_atom_type(i) == offchrg_type
            or molsetup.get_atom_type(j) == offchrg_type
        ):
            continue
        oid_i = indexmap[i]
        oid_j = indexmap[j]
        bonds[oid_i].append("%d" % (oid_j + index_start))
        # bond_orders[oid_i].append("%d" % molsetup.bond_info[(i, j)].order)
        if not molsetup.bond_info[(i, j)].rotatable:
            static_links.append("%d,%d" % (oid_i + index_start, oid_j + index_start))
    bonds = [",".join(j_list) for j_list in bonds]
    bonds_line = "connectivity = {%s}\n" % ("|".join(bonds))
    # bond_orders = [",".join(orders) for orders in bond_orders]
    # bondorder_line = "bond_order = {%s}\n" % ("|".join(bond_orders))
    staticlinks_line = "static_links = {%s}\n" % ("|".join(static_links))

    output = ""
    output += "[Group: %s]\n" % name
    output += positions_block
    output += charges_line
    output += elements_line
    output += bonds_line
    # output += bondorder_line
    output += staticlinks_line
    output += "number = 1\t\t// can only be 1 for the sandbox currently (but any number for classical MC)\n"
    output += "group_dipole = 1\t// not relevant for sandbox but classical MC\n"
    output += "rand_independent=0\t// not relevant for sandbox but classical MC\n"
    output += "bond_range = 4\t\t// bond range AD default\n"
    output += "\n"
    output += get_dihedrals_block(molsetup, indexmap, name)

    return output, indexmap


def get_dihedrals_block(molsetup, indexmap, name):
    # molsetup.dihedral_interactions    is a list of unique fourier_series
    # molsetup.dihedral_partaking_atoms has tuples of atom indices as keys, and the values
    #                                   are the indices in molsetup.dihedral_interactions
    # molsetup.dihedral_labels          also has tuples of atom indices as keys, but the
    #                                   values are not guaranteed to be unique

    # Let's carefully use dihedral_labels to name the interactions
    label_by_index = {}
    atomidx_by_index = {}
    for atomidx in molsetup.dihedral_partaking_atoms:
        a, b, c, d = atomidx
        if (
            molsetup.get_is_ignore(a)
            or molsetup.get_is_ignore(b)
            or molsetup.get_is_ignore(c)
            or molsetup.get_is_ignore(d)
        ):
            continue
        bond_id = Bond.get_bond_id(b, c)
        if not molsetup.bond_info[bond_id].rotatable:
            continue
        index = molsetup.dihedral_partaking_atoms[atomidx]
        atomidx_by_index.setdefault(index, set())
        atomidx_by_index[index].add(atomidx)
        label = (
            molsetup.dihedral_labels[atomidx]
            if atomidx in molsetup.dihedral_labels
            else None
        )
        if label is None:
            label = "from_meeko_%d" % index
        label_by_index.setdefault(index, set())
        label_by_index[index].add(label)
    spent_labels = set()
    for index in label_by_index:
        label = "_".join(label_by_index[index])
        number = 0
        while label in spent_labels:
            number += 1
            label = "_".join(label_by_index[index]) + "_v%d" % number
        label_by_index[index] = label
        spent_labels.add(label)

    text = ""
    for index in label_by_index:
        text += "[Interaction: %s, %s]\n" % (name, label_by_index[index])
        text += "type = dihedral\n"
        atomidx_strings = []
        for atomidx in atomidx_by_index[index]:
            string = ",".join(["%d" % (indexmap[i] + 1) for i in atomidx])
            atomidx_strings.append(string)
        text += "elements = {%s}\n" % ("|".join(atomidx_strings))
        text += "parameters = %s\n" % _aux_fourier_conversion(
            molsetup.dihedral_interactions[index]
        )
        text += "\n"
    return text


def _aux_fourier_conversion(fourier_series):
    # convert from:
    #   k*(1+cos(n*theta-phase))
    # to:
    #   (k/2)*(1+cos(n*(theta+phase)))
    # where n = periodicity
    max_periodicity = max([fs["periodicity"] for fs in fourier_series])
    tmp = [(0, 0)] * max_periodicity
    for fs in fourier_series:
        i = fs["periodicity"] - 1
        k = 2.0 * fs["k"]
        phase = -1 * fs["phase"]
        tmp[i] = (k, phase)
    strings = []
    periodicity = 0
    for k, phase in tmp:
        periodicity += 1
        k_str = "0"
        if phase == 0:
            phase_str = "0"
        else:
            phase_str = ("%f" % (phase / np.pi)).rstrip("0").rstrip(".") + "*pi"
            if phase_str == "1*pi":
                phase_str = "pi"
            if phase_str == "-1*pi":
                phase_str = "-pi"
            if periodicity != 1:
                phase_str += "/%d" % periodicity
        if k != 0:
            k_str = "%f*4.184/60.221" % (k)
        strings.append("%s,%s" % (k_str, phase_str))
    return "(" + ";".join(strings) + ")"


class PDBQTWriterLegacy:

    @staticmethod
    def _get_pdbinfo_fitting_pdb_chars(pdbinfo):
        """return strings and integers that are guaranteed
        to fit within the designated chars of the PDB format"""

        atom_name = pdbinfo.name
        res_name = pdbinfo.resName
        res_num = pdbinfo.resNum
        chain = pdbinfo.chain
        if len(atom_name) > 4:
            atom_name = atom_name[0:4]
        if len(res_name) > 3:
            res_name = res_name[0:3]
        if res_num > 9999:
            res_num = res_num % 10000
        if len(chain) > 1:
            chain = chain[0:1]
        return atom_name, res_name, res_num, chain

    @classmethod
    def _make_pdbqt_line_from_molsetup(cls, setup, atom_idx, count):
        """ """
        pdbinfo = setup.get_pdbinfo(atom_idx)
        if pdbinfo is None:
            pdbinfo = pdbutils.PDBAtomInfo("", "", 0, "")
        atom_name, res_name, res_num, chain = cls._get_pdbinfo_fitting_pdb_chars(
            pdbinfo
        )  # TODO icode
        coord = setup.get_coord(atom_idx)
        atom_type = setup.get_atom_type(atom_idx)
        charge = setup.get_charge(atom_idx)
        pdbqt_line = cls._make_pdbqt_line(
            count, atom_name, res_name, chain, res_num, coord, charge, atom_type
        )
        return pdbqt_line

    @staticmethod
    def _make_pdbqt_line(
        count, atom_name, res_name, chain, res_num, coord, charge, atom_type, icode=""
    ):
        record_type = "ATOM"
        alt_id = " "
        occupancy = 1.0
        temp_factor = 0.0
        atom = "{:6s}{:5d} {:^4s}{:1s}{:3s} {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}    {:6.3f} {:<2s}"
        pdbqt_line = atom.format(
            record_type,
            count,
            atom_name,
            alt_id,
            res_name,
            chain,
            res_num,
            icode,
            float(coord[0]),
            float(coord[1]),
            float(coord[2]),
            occupancy,
            temp_factor,
            charge,
            atom_type,
        )
        return pdbqt_line

    @classmethod
    def _walk_graph_recursive(cls, setup, node, data, edge_start=0, first=False):
        """recursive walk of rigid bodies"""

        if first:
            data["pdbqt_buffer"].append("ROOT")
            member_pool = sorted(setup.flexibility_model["rigid_body_members"][node])
        else:
            member_pool = setup.flexibility_model["rigid_body_members"][node][:]
            member_pool.remove(edge_start)
            member_pool = [edge_start] + member_pool

        for member in member_pool:
            if setup.get_is_ignore(member) == 1:
                continue
            pdbqt_line = cls._make_pdbqt_line_from_molsetup(
                setup, member, data["count"]
            )
            data["pdbqt_buffer"].append(pdbqt_line)
            data["numbering"][member] = data["count"]  # count starts at 1
            data["count"] += 1

        if first:
            data["pdbqt_buffer"].append("ENDROOT")

        data["visited"].append(node)

        for neigh in setup.flexibility_model["rigid_body_graph"][node]:
            if neigh in data["visited"]:
                continue

            # Write the branch
            begin, next_index = setup.flexibility_model["rigid_body_connectivity"][
                node, neigh
            ]

            # do not write branch (or anything downstream) if any of the two atoms
            # defining the rotatable bond are ignored
            if setup.get_is_ignore(begin) or setup.get_is_ignore(next_index):
                continue

            begin = data["numbering"][begin]
            end = data["count"]

            data["pdbqt_buffer"].append("BRANCH %3d %3d" % (begin, end))
            data = cls._walk_graph_recursive(setup, neigh, data, edge_start=next_index)
            data["pdbqt_buffer"].append("ENDBRANCH %3d %3d" % (begin, end))

        return data

    @staticmethod
    def _is_molsetup_ok(setup, bad_charge_ok):

        success = True
        error_msg = ""

        if len(setup.restraints):
            error_msg = "molsetup has restraints but these can't be written to PDBQT"
            success = False

        for atom in setup.atoms:
            if atom.is_ignore:
                continue
            if atom.atom_type is None:
                error_msg += "atom number %d has None type, mol name: %s\n" % (
                    atom.index,
                    setup.get_mol_name(),
                )
                success = False
        for atom in setup.atoms:
            if atom.is_ignore:
                continue
            if atom.atom_type is None:
                error_msg += "atom number %d has None type, mol name: %s\n" % (
                    atom.index,
                    setup.get_mol_name(),
                )
                success = False
            c = atom.charge
            if not bad_charge_ok and (
                type(c) != float and type(c) != int or math.isnan(c) or math.isinf(c)
            ):
                error_msg += (
                    "atom number %d has non finite charge, mol name: %s, charge: %s\n"
                    % (atom.index, setup.get_mol_name(), str(c))
                )
                success = False

        return success, error_msg

    @classmethod
    def write_string_from_polymer(cls, polymer):
        rigid_pdbqt_string, flex_pdbqt_dict = cls.write_from_polymer(
            polymer
        )
        flex_pdbqt_string = ""
        for res_id, pdbqt_string in flex_pdbqt_dict.items():
            flex_pdbqt_string += pdbqt_string
        return rigid_pdbqt_string, flex_pdbqt_string

    @classmethod
    def write_from_polymer(cls, polymer):
        rigid_pdbqt_string = ""
        flex_pdbqt_dict = {}
        atom_count = 0
        flex_atom_count = 0
        for res_id, monomer in polymer.get_valid_monomers().items():
            chain, resnum = res_id.split(":")
            if resnum[-1].isalpha():
                icode = resnum[-1]
                resnum = int(resnum[:-1])
            else:
                icode = ""
                resnum = int(resnum)
            molsetup = monomer.molsetup
            resname = monomer.input_resname
            if monomer.is_movable:
                original_ignore = {atom.index: atom.is_ignore for atom in molsetup.atoms}
                graph = molsetup.flexibility_model["rigid_body_graph"]
                root = molsetup.flexibility_model["root"]
                if len(graph[root]) != 1:
                    raise RuntimeError(
                        f"flexible residue {res_id} has {len(graph[root])}"
                        " rotatable bonds from root, but PDBQT is limited to 1"
                    )
                # set ignore to True for static atoms of flexible sidechains
                # to exclude them from the PDBQT string
                for atom_idx, is_flex in enumerate(monomer.is_flexres_atom):
                        molsetup.atoms[atom_idx].is_ignore = not is_flex
                this_flex_pdbqt, ok, err = PDBQTWriterLegacy.write_string(
                    molsetup, remove_smiles=True, add_index_map=True
                )
                for atom in molsetup.atoms:
                    atom.is_ignore = original_ignore[atom.index]
                if not ok:
                    raise RuntimeError(err)
                this_flex_pdbqt, flex_atom_count = (
                    cls.adapt_pdbqt_for_autodock4_flexres(
                        this_flex_pdbqt,
                        resname,
                        chain,
                        int(resnum),
                        skip_rename_ca_cb=True,
                        atom_count=flex_atom_count,
                    )
                )
                flex_pdbqt_dict[res_id] = this_flex_pdbqt

            for atom_idx, atom in enumerate(molsetup.atoms):
                if atom.is_ignore or monomer.is_flexres_atom[atom_idx]:
                    continue
                atom_type = atom.atom_type
                coord = atom.coord
                atom_name = atom.pdbinfo.name
                charge = atom.charge
                atom_count += 1
                rigid_pdbqt_string += (
                    cls._make_pdbqt_line(
                        atom_count,
                        atom_name,
                        resname,
                        chain,
                        resnum,
                        coord,
                        charge,
                        atom_type,
                        icode,
                    )
                    + eol
                )
        return rigid_pdbqt_string, flex_pdbqt_dict

    @classmethod
    def write_string(
        cls, setup, add_index_map=False, remove_smiles=False, bad_charge_ok=False
    ):
        """Output a PDBQT file as a string.

        Args:
            setup: RDKitMoleculeSetup

        Returns:
            str:  PDBQT string of the molecule
            bool: success
            str:  error message
        """

        success, error_msg = cls._is_molsetup_ok(setup, bad_charge_ok)
        if not success:
            pdbqt_string = ""
            return pdbqt_string, success, error_msg

        data = {
            "visited": [],
            "numbering": {},
            "pdbqt_buffer": [],
            "count": 1,
        }
        atom_counter = {}

        torsdof = len(setup.flexibility_model["rigid_body_graph"]) - 1

        if "torsions_org" in setup.flexibility_model:
            torsdof_org = setup.flexibility_model["torsions_org"]
            data["pdbqt_buffer"].append(
                "REMARK Flexibility Score: %8.3f" % setup.flexibility_model["score"]
            )
            active_tors = torsdof_org
        else:
            active_tors = torsdof

        data = cls._walk_graph_recursive(
            setup, setup.flexibility_model["root"], data, first=True
        )

        if add_index_map:
            for i, remark_line in enumerate(
                cls.remark_index_map(setup, data["numbering"])
            ):
                # Need to use 'insert' because data["numbering"]
                # is populated in self._walk_graph_recursive.
                data["pdbqt_buffer"].insert(i, remark_line)

        if not remove_smiles:
            smiles, order = setup.get_smiles_and_order()
            missing_h = []  # hydrogens which are not in the smiles
            strings_h_parent = []
            for key in data["numbering"]:
                if setup.atoms[key].is_pseudo_atom:
                    continue
                if key not in order:
                    if setup.get_atomic_num(key) != 1:
                        error_msg += (
                            "non-Hydrogen atom unexpectedely missing from smiles!?"
                        )
                        error_msg += " (mol name: %s)\n" % setup.get_mol_name()
                        pdbqt_string = ""
                        success = False
                        return pdbqt_string, success, error_msg
                    missing_h.append(key)
                    parents = setup.get_neighbors(key)
                    parents = [
                        i for i in parents if i < setup.true_atom_count
                    ]  # exclude pseudos
                    if len(parents) != 1:
                        error_msg += (
                            f"expected hydrogen {key} to be bonded to exactly one atom"
                            f" but it's bonded to {parents}"
                        )
                        error_msg += " (mol name: %s)\n" % setup.get_mol_name()
                        pdbqt_string = ""
                        success = False
                        return pdbqt_string, success, error_msg
                    parent_idx = order[parents[0]]  # already 1-indexed
                    string = " %d %d" % (
                        parent_idx,
                        data["numbering"][key],
                    )  # key 0-indexed; _numbering[key] 1-indexed
                    strings_h_parent.append(string)
            remarks_h_parent = cls.break_long_remark_lines(
                strings_h_parent, "REMARK H PARENT"
            )
            remark_prefix = "REMARK SMILES IDX"
            remark_idxmap = cls.remark_index_map(
                setup, data["numbering"], order, remark_prefix, missing_h
            )
            remarks = []
            remarks.append("REMARK SMILES %s" % smiles)  # break line at 79 chars?
            remarks.extend(remark_idxmap)
            remarks.extend(remarks_h_parent)

            for i, remark_line in enumerate(remarks):
                # Need to use 'insert' because data["numbering"]
                # is populated in self._walk_graph_recursive.
                data["pdbqt_buffer"].insert(i, remark_line)

        # torsdof is always going to be the one of the rigid, non-macrocyclic one
        data["pdbqt_buffer"].append("TORSDOF %d" % active_tors)

        pdbqt_string = eol.join(data["pdbqt_buffer"]) + eol
        return pdbqt_string, success, error_msg

    @classmethod
    def remark_index_map(
        cls, setup, numbering, order=None, prefix="REMARK INDEX MAP", missing_h=()
    ):
        """write mapping of atom indices from input molecule to output PDBQT
        order[ob_index(i.e. 'key')] = smiles_index
        """

        if order is None:
            order = {key: key + 1 for key in numbering}  # key+1 breaks OB
        # max_line_length = 79
        # remark_lines = []
        # line = prefix
        strings = []
        for key in numbering:
            if setup.atoms[key].is_pseudo_atom:
                continue
            if key in missing_h:
                continue
            string = " %d %d" % (order[key], numbering[key])
            strings.append(string)
        return cls.break_long_remark_lines(strings, prefix)
        #    candidate_text = " %d %d" % (order[key], self._numbering[key])
        #    if (len(line) + len(candidate_text)) < max_line_length:
        #        line += candidate_text
        #    else:
        #        remark_lines.append(line)
        #        line = 'REMARK INDEX MAP' + candidate_text
        # remark_lines.append(line)
        # return remark_lines

    @staticmethod
    def break_long_remark_lines(strings, prefix, max_line_length=79):
        remarks = [prefix]
        for string in strings:
            if (len(remarks[-1]) + len(string)) < max_line_length:
                remarks[-1] += string
            else:
                remarks.append(prefix + string)
        return remarks

    @staticmethod
    def adapt_pdbqt_for_autodock4_flexres(
        pdbqt_string, res, chain, num, skip_rename_ca_cb=False, atom_count=None
    ):
        """adapt pdbqt_string to be compatible with AutoDock4 requirements:
         - first and second atoms named CA and CB
         - write BEGIN_RES / END_RES
         - remove TORSDOF
        this is for covalent docking (tethered)
        """
        new_string = "BEGIN_RES %s %s %s" % (res, chain, num) + eol
        atom_number = 0
        offset = atom_count
        for line in pdbqt_string.split(eol):
            if line == "":
                continue
            if line.startswith("TORSDOF"):
                continue
            if line.startswith("ATOM"):
                if not skip_rename_ca_cb:
                    atom_number += 1
                    if atom_number == 1:
                        line = line[:13] + "CA" + line[15:]
                    elif atom_number == 2:
                        line = line[:13] + "CB" + line[15:]
                if atom_count is not None:
                    atom_count += 1
                    n = "%5d" % atom_count
                    n = n[:5]
                    line = line[:6] + n + line[11:]
                new_string += line + eol
                continue
            elif offset is not None and (
                line.startswith("BRANCH") or line.startswith("ENDBRANCH")
            ):
                keyword, i, j = line.split()
                new_string += (
                    f"{keyword} {int(i)+offset:3d} {int(j)+offset:3d}" + eol
                )
                continue
            new_string += line + eol
        new_string += "END_RES %s %s %s" % (res, chain, num) + eol
        if atom_count is None:
            return new_string  # just keeping backwards compatibility
        else:
            return new_string, atom_count



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#

import argparse
from datetime import datetime
import io
import os
from os import linesep as eol
import sys
import json
import tarfile
import warnings

from rdkit import Chem

from meeko import MoleculePreparation
from meeko import rdkitutils
from meeko import PDBQTWriterLegacy

try:
    import prody
    from meeko import CovalentBuilder

    _prody_parsers = {"pdb": prody.parsePDB, "mmcif": prody.parseMMCIF}
except ImportError as err:
    _has_prody = False
    _prody_parsers = {}
    _prody_import_error = err
else:
    _has_prody = True


def cmd_lineparser():
    backend = "rdkit"

    conf_parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False,
    )
    conf_parser.add_argument(
        "-c",
        "--config_file",
        help="configure MoleculePreparation from JSON file. Overriden by command line args.",
    )
    confargs, remaining_argv = conf_parser.parse_known_args()

    config = MoleculePreparation.get_defaults_dict()

    if confargs.config_file is not None:
        with open(confargs.config_file) as f:
            c = json.load(f)
            config.update(c)

    parser = (
        argparse.ArgumentParser()
    )  # parents=[conf_parser]) # parents shows --config_file in help msg
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="print information about molecule setup",
    )

    io_group = parser.add_argument_group("Input/Output")
    io_group.add_argument(
        "-i",
        "--mol",
        dest="input_molecule_filename",
        required=True,
        action="store",
        help="molecule file (MOL2, SDF,...)",
    )
    io_group.add_argument(
        "--name_from_prop",
        help="set molecule name from RDKit/SDF property",
    )
    io_group.add_argument(
        "-o",
        "--out",
        dest="output_pdbqt_filename",
        action="store",
        help="output pdbqt filename. Single molecule input only.",
    )
    io_group.add_argument(
        "--multimol_outdir",
        dest="multimol_output_dir",
        action="store",
        help="folder to write output pdbqt for multi-mol inputs. Incompatible with -o/--out and -/--.",
    )
    io_group.add_argument(
        "--multimol_prefix",
        dest="multimol_prefix",
        action="store",
        help="replace internal molecule name in multi-molecule input by specified prefix. Incompatible with -o/--out and -/--.",
    )
    io_group.add_argument(
        "-z",
        "--multimol_targz",
        action="store_true",
        help="compress output files in .tar.gz",
    )
    io_group.add_argument(
        "--multimol_targz_size",
        type=int,
        default=10000,
        help="number of PDBQT files per .tar.gz",
    )
    io_group.add_argument(
        "-",
        "--",
        dest="redirect_stdout",
        action="store_true",
        help="do not write file, redirect output to STDOUT. Argument -o/--out is ignored. Single molecule input only.",
    )

    config_group = parser.add_argument_group("Molecule preparation")
    config_group.add_argument(
        "-c",
        "--config_file",
        help="configure MoleculePreparation from JSON file. Overriden by command line args.",
    )  # parsed above by conf_parser, here for help msg
    config_group.add_argument(
        "--rigid_macrocycles",
        dest="rigid_macrocycles",
        action="store_true",
        help="keep macrocycles rigid in input conformation",
    )
    config_group.add_argument(
        "--macrocycle_allow_A",
        action="store_true",
        help="allow bond break with atom type A, retyped as C",
    )
    config_group.add_argument(
        "--keep_chorded_rings",
        dest="keep_chorded_rings",
        action="store_true",
        help="return all rings from exhaustive perception",
    )
    config_group.add_argument(
        "--keep_equivalent_rings",
        dest="keep_equivalent_rings",
        action="store_true",
        help="equivalent rings have the same size and neighbors",
    )
    config_group.add_argument(
        "--min_ring_size",
        dest="min_ring_size",
        type=int,
        help="min nr of atoms in ring for opening",
    )
    config_group.add_argument(
        "-w",
        "--hydrate",
        dest="hydrate",
        action="store_true",
        help="add water molecules for hydrated docking",
    )
    config_group.add_argument(
        "--merge_these_atom_types",
        dest="merge_these_atom_types",
        nargs="*",
        help='list of atom types to merge, default is "H"',
        default=["H"],
    )
    config_group.add_argument(
        "-r",
        "--rigidify_bonds_smarts",
        dest="rigidify_bonds_smarts",
        action="append",
        help="SMARTS patterns to rigidify bonds",
        metavar="SMARTS",
    )
    config_group.add_argument(
        "-b",
        "--rigidify_bonds_indices",
        dest="rigidify_bonds_indices",
        action="append",
        help="indices of two atoms (in the SMARTS) that define a bond (start at 1)",
        nargs="+",
        type=int,
        metavar="i j",
    )
    config_group.add_argument(
        "-a",
        "--flexible_amides",
        dest="flexible_amides",
        action="store_true",
        help="allow amide bonds to rotate and be non-planar, which is bad",
    )
    config_group.add_argument(
        "-p",
        "--load_atom_params",
        nargs="+",
        action="store",
        default=["ad4_types"],
        help="filename with SMARTS defined atom types (JSON format)",
    )
    config_group.add_argument(
        "-aa",
        "--add_atom_types",
        dest="add_atom_types_json",
        action="append",
        help="Additional atom types to assign (JSON formated)",
        metavar="[{'smarts': '<smarts pattern>', 'atype': ',atomtype name>'}, {'smarts': '<smarts pattern>', 'atype': ',atomtype name>'}]",
    )
    config_group.add_argument(
        "--double_bond_penalty",
        help="penalty > 100 prevents breaking double bonds",
        type=int,
    )
    config_group.add_argument(
        "--charge_model",
        choices=("gasteiger", "espaloma", "zero"),
        help="default is 'gasteiger', 'zero' sets all zeros",
        default="gasteiger",
    )
    config_group.add_argument(
        "--bad_charge_ok",
        help="NaN and Inf charges allowed in PDBQT",
        action="store_true",
    )
    config_group.add_argument(
        "--add_index_map",
        dest="add_index_map",
        action="store_true",
        help="write map of atom indices from input to pdbqt",
    )
    config_group.add_argument(
        "--remove_smiles",
        dest="remove_smiles",
        action="store_true",
        help="do not write smiles as remark to pdbqt",
    )
    reactive_group = parser.add_argument_group("Reactive docking")
    reactive_group.add_argument(
        "--reactive_smarts", help="SMARTS pattern for reactive group"
    )
    reactive_group.add_argument(
        "--reactive_smarts_idx",
        help="index (1-based) of the reactive atom in --reactive_smarts",
        type=int,
    )

    need_prody_msg = ""
    if not _has_prody:
        need_prody_msg = ". Needs Prody which can be installed from PyPI or conda-forge"
    covalent_group = parser.add_argument_group(
        "Covalent docking (tethered)%s" % (need_prody_msg)
    )
    covalent_group.add_argument(
        "--receptor",
        help="receptor filename. Supported formats: [%s]%s"
        % ("/".join(list(_prody_parsers.keys())), need_prody_msg),
    )
    covalent_group.add_argument(
        "--rec_residue", help='examples: "A:LYS:204", "A:HIS:", ":LYS:"'
    )
    covalent_group.add_argument(
        "--tether_smarts",
        help="SMARTS pattern to define ligand atoms for receptor attachment",
    )
    covalent_group.add_argument(
        "--tether_smarts_indices",
        type=int,
        nargs=2,
        required=False,
        metavar="IDX",
        default=[1, 2],
        help="indices (1-based) of the SMARTS atoms that will be attached (default: 1 2)",
    )

    parser.set_defaults(**config)
    args = parser.parse_args(remaining_argv)

    # check reactive arguments
    if (args.reactive_smarts is None) != (args.reactive_smarts_idx is None):
        print(
            "Arguments --reactive_smarts and --reactive_smarts_idx require each other",
            file=sys.stderr,
        )
        sys.exit(2)
    elif args.reactive_smarts_idx is not None:
        if args.reactive_smarts_idx < 1:
            print(
                "--reactive_smarts_idx is 1-indexed, but got %d"
                % args.reactive_smarts_idx,
                file=sys.stderr,
            )
            sys.exit(2)
        args.reactive_smarts_idx -= 1  # convert from 1- to 0-index

    # command line arguments override config
    for key in config:
        if key in args.__dict__:
            config[key] = args.__dict__[key]

    config["load_atom_params"] = args.load_atom_params

    if args.add_atom_types_json is not None:
        additional_ats = []
        for at in args.add_atom_types_json:
            at = json.loads(at)
            if type(at) == dict:
                additional_ats.append(at)
            elif type(at) == list:
                additional_ats.extend(at)
        config["add_atom_types"] = additional_ats

    if args.multimol_output_dir is not None or args.multimol_prefix is not None:
        if args.output_pdbqt_filename is not None:
            print(
                "Warning: -o/--out ignored with --multimol_outdir or --multimol_prefix",
                file=sys.stderr,
            )
        if args.redirect_stdout:
            print(
                "Warning: -/-- ignored with --multimol_outdir or --multimol_prefix",
                file=sys.stderr,
            )

    # verify sanity of covalent docking input
    num_required_covalent_args = 0
    num_required_covalent_args += int(args.receptor is not None)
    num_required_covalent_args += int(args.rec_residue is not None)
    num_required_covalent_args += int(args.tether_smarts is not None)
    if num_required_covalent_args not in [0, 3]:
        print(
            "Error: --receptor, --rec_residue, and --tether_smarts are all required for covalent docking."
        )
        sys.exit(2)
    is_covalent = num_required_covalent_args == 3
    if is_covalent and not _has_prody:
        msg = "Covalent docking requires Prody which is not installed." + eol
        msg += "Installable from PyPI (pip install prody) or conda-forge (micromamba install prody)"
        print(_prody_import_error, file=sys.stderr)
        print(msg)
        sys.exit(2)
    if min(args.tether_smarts_indices) < 1:
        print(
            "--tether_smarts_indices is 1-indexed, all values must be greater than zero",
            file=sys.stderr,
        )
        sys.exit(2)
    args.tether_smarts_indices = [
        i - 1 for i in args.tether_smarts_indices
    ]  # convert to 0-index

    # verify sanity of SMARTS patterns to make bonds rigid and convert to 0-based indices
    rigidify_bonds_smarts = config["rigidify_bonds_smarts"]
    rigidify_bonds_indices = config["rigidify_bonds_indices"]
    if len(rigidify_bonds_indices) != len(rigidify_bonds_smarts):
        raise RuntimeError(
            "length of --rigidify_bonds_indices differs from length of --rigidify_bonds_smarts"
        )
    for indices in rigidify_bonds_indices:
        if len(indices) != 2:
            raise RuntimeError(
                "--rigidify_bonds_indices must specify pairs, e.g. -b 1 2 -b 3 4"
            )
        indices[0] = indices[0] - 1  # convert from 1- to 0-index
        indices[1] = indices[1] - 1

    return args, config, backend, is_covalent


class Output:
    def __init__(self, multimol_output_dir, multimol_targz, multimol_targz_size, multimol_prefix, redirect_stdout, output_filename, name_from_prop):
        is_multimol = (multimol_prefix is not None) or (multimol_output_dir is not None)
        self._mkdir(multimol_output_dir)

        if multimol_output_dir is None:
            multimol_output_dir = "."
        self.multimol_output_dir = multimol_output_dir
        self.multimol_targz = multimol_targz
        self.multimol_targz_size = max(int(multimol_targz_size), 1)
        self.tarf = None
        self.tar_pdbqt_count = 0
        self.tarf_index = 0
        self.multimol_prefix = multimol_prefix
        self.redirect_stdout = redirect_stdout
        self.output_filename = output_filename
        self.is_multimol = is_multimol
        self.visited_filenames = set()
        self.duplicate_filenames = set()
        self.visited_names = set()
        self.duplicate_names = set()
        self.num_files_written = 0
        self.counter = 0
        self.name_from_prop = name_from_prop

    def _open_new_tar(self):
        self.tarf_index += 1
        prefix = "" if self.multimol_prefix is None else self.multimol_prefix
        tgz_path = os.path.join(
            self.multimol_output_dir,
            f"{prefix}{self.tarf_index:07d}.tar.gz"
        )
        if self.tarf is not None:
            self.tarf.close()
        tarf = tarfile.open(tgz_path, "w:gz")
        return tarf

    def _add_to_tar(self, pdbqt_string, name):
        if self.tarf is None or self.tar_pdbqt_count >= self.multimol_targz_size:
            self.tarf = self._open_new_tar()
            self.tar_pdbqt_count = 0
        tarinfo = tarfile.TarInfo(name=f"{name}.pdbqt")
        tarinfo.size = len(pdbqt_string)
        tarinfo.mtime = datetime.timestamp(datetime.now())
        self.tarf.addfile(tarinfo, io.BytesIO(pdbqt_string.encode()))
        self.tar_pdbqt_count += 1
        return

    def __call__(self, pdbqt_string, name, suffixes=()):
        self.counter += 1
        if name in self.visited_names:
            self.duplicate_names.add(name)
        self.visited_names.add(name)
        if self.multimol_prefix is not None:
            name = "%s-%d" % (self.multimol_prefix, self.counter)
        for suffix in suffixes:
            if suffix is not None and len(suffix) > 0:
                name += "_" + suffix

        if self.is_multimol:
            if name in self.visited_filenames:
                self.duplicate_filenames.add(name)
                repeat_id = 1
                newname = name + "-again%d" % repeat_id
                while newname in self.visited_filenames:
                    repeat_id += 1
                    newname = name + "-again%d" % repeat_id
                print(
                    "Renaming %s to %s to disambiguate filename" % (name, newname),
                    file=sys.stderr,
                )
                name = newname

            self.visited_filenames.add(name)

            if self.multimol_targz:
                self._add_to_tar(pdbqt_string, name)
            else:
                fpath = os.path.join(self.multimol_output_dir, name + ".pdbqt")
                with open(fpath, "w") as f:
                    print(pdbqt_string, end="", file=f)
            self.num_files_written += 1

        elif self.redirect_stdout:
            print(pdbqt_string, end="")
        else:
            if self.output_filename is None:
                filename = "%s.pdbqt" % name
            else:
                filename = self.output_filename
            with open(filename, "w") as f:
                print(pdbqt_string, end="", file=f)
            self.num_files_written += 1

    def _mkdir(self, multimol_output_dir):
        """make directory if it doesn't exist yet"""
        if multimol_output_dir is not None:
            if not os.path.exists(multimol_output_dir):
                os.mkdir(multimol_output_dir)

    def get_duplicates_info_string(self):
        if not self.is_multimol:
            return None
        if len(self.duplicate_names):
            # with multimol_prefix with can have duplicate molecule names,
            # but not duplicate filenames. This warning is for such cases.
            string = "Warning: %d molecules have duplicated names" % len(
                self.duplicate_names
            )
        else:
            string = "No duplicate molecule molecule names were found"
        # if we have duplicate_filenames, we also have duplicate molecule names,
        # but it suffices to return the string below
        if len(self.duplicate_filenames):
            string = (
                "Warning: %d molecules with repeated names were suffixed with -again<n>"
                % len(self.duplicate_filenames)
            )
        return string

    @staticmethod
    def get_suffixes(molsetups):
        if len(molsetups) == 1:
            return ("",)  # no suffix needed
        else:
            return tuple("mk%d" % (i + 1) for i in range(len(molsetups)))

def main():
    args, config, backend, is_covalent = cmd_lineparser()
    input_molecule_filename = args.input_molecule_filename

    # read input
    input_fname, ext = os.path.splitext(input_molecule_filename)
    ext = ext[1:].lower()
    if backend == "rdkit":
        parsers = {
            "sdf": Chem.SDMolSupplier,
            "mol2": rdkitutils.Mol2MolSupplier,
            "mol": Chem.SDMolSupplier,
        }
        if not ext in parsers:
            print(
                "*ERROR* Format [%s] not in supported formats [%s]"
                % (ext, "/".join(list(parsers.keys())))
            )
            sys.exit(1)
        mol_supplier = parsers[ext](
            input_molecule_filename, removeHs=False
        )  # input must have explicit H
    
    # configure output writer
    if args.output_pdbqt_filename is None:
        output_filename = input_fname + ".pdbqt"
    else:
        output_filename = args.output_pdbqt_filename


    output = Output(
        args.multimol_output_dir,
        args.multimol_targz,
        args.multimol_targz_size,
        args.multimol_prefix,
        args.redirect_stdout,
        output_filename,
        args.name_from_prop,
    )

    # initialize covalent object for receptor
    if is_covalent:
        rec_filename = args.receptor
        _, rec_extension = os.path.splitext(rec_filename)
        rec_extension = rec_extension[1:].lower()
        prody_parser = _prody_parsers[rec_extension]
        rec_prody_mol = prody_parser(rec_filename)
        covalent_builder = CovalentBuilder(rec_prody_mol, args.rec_residue)

    input_mol_counter = 0
    input_mol_skipped = 0
    input_mol_with_failure = (
        0  # if reactive or covalent, each mol can yield multiple PDBQT
    )
    nr_failures = 0
    is_after_first = False
    preparator = MoleculePreparation.from_config(config)
    for mol in mol_supplier:
        if is_after_first and output.is_multimol == False:
            print("Processed only the first molecule of multiple molecule input.")
            print(
                "Use --multimol_prefix and/or --multimol_outdir to process all molecules in %s."
                % (input_molecule_filename)
            )
            break

        # check that molecule was successfully loaded
        if backend == "rdkit":
            is_valid = mol is not None
        input_mol_skipped += int(is_valid == False)
        if not is_valid:
            continue

        this_mol_had_failure = False

        if args.name_from_prop is not None:
            if mol.HasProp(args.name_from_prop):
                name = mol.GetProp(args.name_from_prop)
            else:
                continue  # TODO log this event
        else:
            name = mol.GetProp("_Name")
        is_after_first = True

        if is_covalent:
            for cov_lig in covalent_builder.process(
                mol, args.tether_smarts, args.tether_smarts_indices
            ):
                root_atom_index = cov_lig.indices[0]
                molsetups = preparator.prepare(
                    cov_lig.mol,
                    root_atom_index=root_atom_index,
                    not_terminal_atoms=[root_atom_index],
                )
                chain, res, num = cov_lig.res_id
                suffixes = output.get_suffixes(molsetups)
                for molsetup, suffix in zip(molsetups, suffixes):
                    pdbqt_string, success, error_msg = PDBQTWriterLegacy.write_string(
                        molsetup,
                        bad_charge_ok=args.bad_charge_ok,
                        add_index_map=args.add_index_map,
                    )
                    if success:
                        pdbqt_string = (
                            PDBQTWriterLegacy.adapt_pdbqt_for_autodock4_flexres(
                                pdbqt_string, res, chain, num
                            )
                        )
                        name = molsetup.name
                        output(pdbqt_string, name, (cov_lig.label, suffix))
                    else:
                        nr_failures += 1
                        this_mol_had_failure = True
                        print(error_msg, file=sys.stderr)

        else:
            molsetups = preparator.prepare(mol)
            if len(molsetups) > 1:
                output.is_multimol = True
            suffixes = output.get_suffixes(molsetups)
            for molsetup, suffix in zip(molsetups, suffixes):
                pdbqt_string, success, error_msg = PDBQTWriterLegacy.write_string(
                    molsetup,
                    bad_charge_ok=args.bad_charge_ok,
                    add_index_map=args.add_index_map,
                )
                if success:
                    output(pdbqt_string, name, (suffix,))
                    if args.verbose:
                        molsetup.show()
                else:
                    nr_failures += 1
                    this_mol_had_failure = True
                    print(error_msg, file=sys.stderr)

        input_mol_with_failure += int(this_mol_had_failure)

    # Close the last tarf opened
    if output.tarf is not None:
        output.tarf.close()

    # Print final status
    print(
        "Input molecules processed: %d, skipped: %d"
        % (output.counter, input_mol_skipped)
    )
    print("PDBQT files written: %d" % (output.num_files_written))
    print("PDBQT files not written due to error: %d" % (nr_failures))
    print("Input molecules with errors: %d" % (input_mol_with_failure))
    if output.is_multimol:
        # would be None if not is_multimol
        print(output.get_duplicates_info_string())

    # Determine if exit code should be non-zero based on processing results
    if output.num_files_written == 0 and not output.redirect_stdout:
        print("No PDBQT files were written due to errors!")
        sys.exit(3)  # full failure
    elif input_mol_with_failure > 0:
        print("Some molecules encountered errors!")
        sys.exit(4)  # partial failure
    return

if __name__ == '__main__':
    sys.exit(main())



import json
import pathlib
from meeko import utils
from meeko import Polymer
from meeko import MoleculePreparation
from meeko import ResidueChemTemplates
from meeko import PDBQTMolecule
from meeko import export_pdb_updated_flexres
import meeko

pkgdir = pathlib.Path(meeko.__file__).parents[1]
just_three_residues = pkgdir / "test/polymer_data/just-three-residues.pdb"
j3r_docked = pkgdir / "test/polymer_data/just-three-residues_vina_flexres.pdbqt"
j3r_idx_docked = pkgdir / "test/polymer_data/just-three-residues_vina_flexres_idxmap.pdbqt"

chem_templates = ResidueChemTemplates.create_from_defaults()
mk_prep = MoleculePreparation(
    merge_these_atom_types=["H"],
    charge_model="gasteiger",
    load_atom_params="ad4_types",
)

def test_begin_res_parsing():
    assert utils.parse_begin_res("SE8 C 23") == "C:23" 
    assert utils.parse_begin_res("SE8  23") == ":23"
    assert utils.parse_begin_res("SE8  23A") == ":23A"
    assert utils.parse_begin_res("SER A1234A") == "A:1234A"
    assert utils.parse_begin_res("    A 999") == "A:999"
    assert utils.parse_begin_res("  1") == ":1"
    assert utils.parse_begin_res(" A1234A") == "A:1234A"
    assert utils.parse_begin_res(" B1234") == "B:1234"
    assert utils.parse_begin_res("S  23A") == "S:23A"

def get_x_from_pdb_str(pdbstr, wanted_chain, wanted_resnum, wanted_name):
    for line in pdbstr.splitlines():
        if not line.startswith("ATOM") and not line.startswith("HETATM"):
            continue
        resn = int(line[22:26])
        name = line[12:16].strip()
        c = line[21].strip()
        if resn == wanted_resnum and name == wanted_name and c == wanted_chain:
            return float(line[30:38])
    return None

def test_export_sidechains_no_idxmap():
    with open(just_three_residues) as f:
        pdb_string = f.read()
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
    )
    # docking had these two flexible
    polymer.flexibilize_sidechain(":15", mk_prep)
    polymer.flexibilize_sidechain(":16", mk_prep)
    
    with open(j3r_docked) as f:
        docked_pdbqt_string = f.read()
    pdbqt_mol = PDBQTMolecule(docked_pdbqt_string, skip_typing=True)
    pdbqt_mol._current_pose = 0
    pdb_string = export_pdb_updated_flexres(polymer, pdbqt_mol)
    x = get_x_from_pdb_str(pdb_string, "", 15, "SD")
    assert abs(x - 10.529) < 0.0001
    pdbqt_mol._current_pose = 1
    pdb_string = export_pdb_updated_flexres(polymer, pdbqt_mol)
    x = get_x_from_pdb_str(pdb_string, "", 16, "HG")
    assert abs(x - 18.724) < 0.0001

    # rebuilding polymer can fail if structure is mangled
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
    )
    polymer.flexibilize_sidechain(":15", mk_prep)
    polymer.flexibilize_sidechain(":16", mk_prep)
    assert len(polymer.get_valid_monomers()) == 3

    # with INDEX MAP in flexres PDBQT
    with open(j3r_idx_docked) as f:
        docked_pdbqt_string = f.read()
    pdbqt_mol = PDBQTMolecule(docked_pdbqt_string, skip_typing=True)
    pdbqt_mol._current_pose = 0
    pdb_string = export_pdb_updated_flexres(polymer, pdbqt_mol)
    x = get_x_from_pdb_str(pdb_string, "", 15, "SD")
    assert abs(x - 10.577) < 0.0001
    pdbqt_mol._current_pose = 1
    pdb_string = export_pdb_updated_flexres(polymer, pdbqt_mol)
    x = get_x_from_pdb_str(pdb_string, "", 16, "HG")
    assert abs(x - 18.690) < 0.0001

    # rebuilding polymer can fail if structure is mangled
    polymer = Polymer.from_pdb_string(
        pdb_string,
        chem_templates,
        mk_prep,
    )
    polymer.flexibilize_sidechain(":16", mk_prep)
    assert len(polymer.get_valid_monomers()) == 3



from rdkit import Chem
from rdkit.Chem import rdDetermineBonds
from rdkit.Geometry import Point3D
from .utils.utils import parse_begin_res
from .utils.utils import mini_periodic_table
from .rdkit_mol_create import RDKitMolCreate
from .polymer import Polymer


mini_periodic_table = {v: k for k, v in mini_periodic_table.items()}

def sidechain_to_mol(pdbqt_atoms):
    positions = []
    mol = Chem.EditableMol(Chem.Mol())
    mol.BeginBatchEdit()
    for row in pdbqt_atoms:
        # the first character of the PDB atom name seems to be the most
        # reliable way to get the element from PDBQT files written by
        # meeko or mgltools
        if len(row["name"]) > 1 and row["name"][0].isdecimal():
            element = row["name"][1]
        elif len(row["name"]) > 0:
            element = row["name"][0] 
        atomic_nr = mini_periodic_table[element]
        atom = Chem.Atom(atomic_nr)
        mol.AddAtom(atom)
        x, y, z = row["xyz"]
        positions.append(Point3D(float(x), float(y), float(z)))
    mol.CommitBatchEdit()
    mol = mol.GetMol()
    conformer = Chem.Conformer(mol.GetNumAtoms())
    for index, position in enumerate(positions):
        conformer.SetAtomPosition(index, position)
    mol.AddConformer(conformer, assignId=True)
    rdDetermineBonds.DetermineConnectivity(mol)
    Chem.SanitizeMol(mol)
    return mol

def export_pdb_updated_flexres(polymer, pdbqt_mol):
    flexres_id = pdbqt_mol._pose_data["mol_index_to_flexible_residue"]
    new_positions = {}
    for mol_idx, atom_idxs in pdbqt_mol._atom_annotations["mol_index"].items():
        if flexres_id[mol_idx] is not None:
            res_id = parse_begin_res(flexres_id[mol_idx])
            atoms = pdbqt_mol.atoms(atom_idxs)
            molsetup_to_pdbqt = pdbqt_mol._pose_data["index_map"][mol_idx]
            
            molsetup_to_template = polymer.monomers[res_id].molsetup_mapidx
            template_to_molsetup = {j: i for i, j in molsetup_to_template.items()}
            
            # use smiles and smiles_index_map in PDBQT REMARK
            # smiles will be None if it's a typical flexres
            # but there will be a valid Smiles string if it's a covalent flexres
            if pdbqt_mol._pose_data["smiles"][mol_idx] is not None: 

                orig_resmol = Chem.Mol(polymer.monomers[res_id].rdkit_mol)
                orig_atomnames = polymer.monomers[res_id].atom_names
                
                backbone_SMARTS = "[NX3]([H])[CX4][CX3](=O)"
                expected_names = ('N', 'H', 'CA', 'C', 'O')

                backbone_qmol = Chem.MolFromSmarts(backbone_SMARTS)
                backbone_matches = orig_resmol.GetSubstructMatches(backbone_qmol)

                if not backbone_matches: 
                    raise RuntimeError(f"Could not find standard backbone structures in receptor {res_id}. ")
                backbone_expected = tuple(orig_atomnames.index(atom_name) for atom_name in expected_names)

                if backbone_expected not in backbone_matches: 
                    raise RuntimeError(f"Could not confirm residue's backbone by atom names in receptor {res_id}. ")
                
                not_backbone = [atom.GetIdx() for atom in orig_resmol.GetAtoms() if atom.GetIdx() not in backbone_expected]
                orig_backbone = Chem.RWMol(orig_resmol)
                for idx in sorted(not_backbone, reverse=True):
                    orig_backbone.RemoveAtom(idx)
                orig_backbone = orig_backbone.GetMol()
              
                covres_mol = RDKitMolCreate.from_pdbqt_mol(pdbqt_mol, 
                                                           only_cluster_leads=False, 
                                                           keep_flexres=True)[mol_idx]
                
                covres_conformer = covres_mol.GetConformer()
                CA_index = None
                target_coord = orig_backbone.GetConformer().GetPositions()[expected_names.index('CA')]
                tolerance = 5e-3
                for atom in covres_mol.GetAtoms():
                    idx = atom.GetIdx()
                    pos = covres_conformer.GetAtomPosition(idx)
                    if (abs(pos.x - target_coord[0]) <= tolerance and
                        abs(pos.y - target_coord[1]) <= tolerance and
                        abs(pos.z - target_coord[2]) <= tolerance):
                        CA_index = idx
                        break
                if CA_index is None:
                    raise RuntimeError(f"Could not determine CA by coordinates in docked pose of {res_id}. ")
                
                CA_in_covres = covres_mol.GetAtomWithIdx(CA_index)
                bonds_to_recover = []
                for bond in CA_in_covres.GetBonds():
                    neighbor_idx = bond.GetOtherAtomIdx(CA_index)
                    bond_type = bond.GetBondType()
                    bonds_to_recover.append((neighbor_idx, bond_type))

                combined_res = Chem.RWMol(Chem.CombineMols(covres_mol, orig_backbone))
                for neighbor_idx, bond_type in bonds_to_recover: 
                    new_CA_idx = covres_mol.GetNumAtoms() + expected_names.index('CA')
                    combined_res.AddBond(new_CA_idx, neighbor_idx, bond_type)
                combined_res.RemoveAtom(CA_index)
                combined_res = combined_res.GetMol()

                polymer.monomers[res_id].rdkit_mol = combined_res
                polymer.monomers[res_id].atom_names = [atom.GetSymbol() for atom in combined_res.GetAtoms()]
                new_positions[res_id] = {idx: coord for idx,coord in enumerate(combined_res.GetConformer().GetPositions())}

            else: # use templates
            
                # use index_map stored in PDBQT REMARK
                if molsetup_to_pdbqt:
                    template_to_pdbqt = {}
                    for i, j in molsetup_to_pdbqt.items():
                        template_to_pdbqt[molsetup_to_template[i - 1]] = j - 1


                # use polymer template to match sidechain
                else:
                    mol = sidechain_to_mol(atoms)
                    key = polymer.monomers[res_id].residue_template_key
                    template = polymer.residue_chem_templates.residue_templates[key]
                    _, template_to_pdbqt = template.match(mol)

                sidechain_positions = {}
                molsetup_matched = set()
                for i, j in template_to_pdbqt.items():
                    molsetup_matched.add(template_to_molsetup[i])
                    sidechain_positions[i] = tuple(atoms["xyz"][j])
                if len(molsetup_matched) != len(template_to_pdbqt):
                    raise RuntimeError(f"{len(molsetup_matched)} {len(template_to_pdbqt)=}")
                is_flexres_atom = polymer.monomers[res_id].is_flexres_atom 
                hit_count = sum([is_flexres_atom[i] for i in molsetup_matched])
                if hit_count != len(molsetup_matched):
                    raise RuntimeError(f"{hit_count=} {len(molsetup_matched)=}")
                if hit_count != sum(is_flexres_atom):
                    raise RuntimeError(f"{hit_count=} {sum(is_flexres_atom)=}")
                # new_positions[res_id] = sidechain_positions

                # remove root atom(s) (often C-alpha) and first atom after bond
                flex_model = polymer.monomers[res_id].molsetup.flexibility_model
                root_body_idx = flex_model["root"]
                graph = flex_model["rigid_body_graph"]
                conn = flex_model["rigid_body_connectivity"]
                rigid_index_by_atom = flex_model["rigid_index_by_atom"]
                first_after_root = set()
                for other_body_idx in graph[root_body_idx]:
                    first_after_root.add(conn[(root_body_idx, other_body_idx)][1])
                to_pop = set()
                for index in sidechain_positions:
                    index_molsetup = template_to_molsetup[index]
                    if (
                        rigid_index_by_atom[index_molsetup] == root_body_idx or 
                        index_molsetup in first_after_root
                    ): 
                        to_pop.add(index) 
                for index in to_pop:
                    sidechain_positions.pop(index)
                new_positions[res_id] = sidechain_positions
    
    pdbstr = polymer.to_pdb(new_positions)
    return pdbstr

def pdb_updated_flexres_from_rdkit(polymer:Polymer, flexres_rdkit_mols:dict):
    """Take dict of flexible residue RDKit molecules and update the polymer PDB with their positions

    Args:
        polymer (Polymer): receptor Polymer to edit
        flexres_rdkit (dict): Dict of RDKit molecule objects of updated flexible residue positions. Key is residue ID number

    Returns:
        str: PDB string for receptor with updated flexres positons
    """
    # TODO: make this a helper function callable by the function above?

    new_positions = {}
    for res_id, mol in flexres_rdkit_mols.items():
        print(Chem.MolToSmiles(mol))
        #mol = Chem.RemoveHs(mol)
        # get templates for matching indices of rdkit mol to monomer in polymer
        key = polymer.monomers[res_id].residue_template_key
        template = polymer.residue_chem_templates.residue_templates[key]
        _, template_to_pdbqt = template.match(mol)
        molsetup_to_template = polymer.monomers[res_id].molsetup_mapidx
        template_to_molsetup = {j: i for i, j in molsetup_to_template.items()}

        sidechain_positions = {}
        molsetup_matched = set()
        matched_but_ignored = 0
        for i, j in template_to_pdbqt.items():
            index_molsetup = template_to_molsetup[i]
            if polymer.monomers[res_id].molsetup.atoms[index_molsetup].is_ignore:  # do not include ignored atoms (non-polar hydrogens)
                matched_but_ignored += 1
            else:
                molsetup_matched.add(template_to_molsetup[i])
                sidechain_positions[i] = mol.GetConformer().GetAtomPosition(j)
        if len(molsetup_matched) != len(template_to_pdbqt) - matched_but_ignored:
            raise RuntimeError(f"{len(molsetup_matched)} {len(template_to_pdbqt)=}")
        is_flexres_atom = polymer.monomers[res_id].is_flexres_atom
        hit_count = sum([is_flexres_atom[i] for i in molsetup_matched])
        if hit_count != len(molsetup_matched):
            raise RuntimeError(f"{hit_count=} {len(molsetup_matched)=}")
        if hit_count != sum(is_flexres_atom):
            raise RuntimeError(f"{hit_count=} {sum(is_flexres_atom)=}")
        # new_positions[res_id] = sidechain_positions

        # remove root atom(s) (often C-alpha) and first atom after bond
        flex_model = polymer.monomers[res_id].molsetup.flexibility_model
        root_body_idx = flex_model["root"]
        graph = flex_model["rigid_body_graph"]
        conn = flex_model["rigid_body_connectivity"]
        rigid_index_by_atom = flex_model["rigid_index_by_atom"]
        first_after_root = set()
        for other_body_idx in graph[root_body_idx]:
            first_after_root.add(conn[(root_body_idx, other_body_idx)][1])
        to_pop = set()
        for index in sidechain_positions:
            print(index)
            index_molsetup = template_to_molsetup[index]
            if (
                rigid_index_by_atom[index_molsetup] == root_body_idx or 
                index_molsetup in first_after_root
            ): 
                to_pop.add(index) 
        for index in to_pop:
            sidechain_positions.pop(index)
        new_positions[res_id] = sidechain_positions

    pdbstr = polymer.to_pdb(new_positions=new_positions)
    return pdbstr



import math
import pathlib
import xml.etree.ElementTree as ET

from rdkit import Chem

from .utils.utils import mini_periodic_table


def load_openff():
    import openforcefields

    p = pathlib.Path(openforcefields.__file__)  # find openff-forcefields xml files
    offxml = p.parents[0] / "offxml" / "openff-2.0.0.offxml"
    offxml = offxml.resolve()
    vdw_list, dihedral_list, vdw_by_type = parse_offxml(offxml)
    return vdw_list, dihedral_list, vdw_by_type


def get_openff_epsilon_sigma(rdmol, vdw_list, vdw_by_type, output_index_start=0):
    from .preparation import MoleculePreparation

    data = {}
    meeko_config = {
        "keep_nonpolar_hydrogens": True,
        "atom_type_smarts": {
            "OFFATOMS": {},
            "ATOM_PARAMS": {"openff-2.0.0": vdw_list},
            "CHARGE_MODEL": "gasteiger",
        },
    }
    mk_prep = MoleculePreparation.from_config(meeko_config)
    mk_prep.prepare(rdmol)
    molsetup = mk_prep.setup
    for atom in molsetup.atoms:
        data[atom.index + output_index_start] = {
            "atom_type": atom.atom_type,
            "epsilon": vdw_by_type[atom.atom_type]["epsilon"],
            "sigma": vdw_by_type[atom.atom_type]["rmin_half"] * 2.0 / (4 ** (1.0 / 12)),
            "gasteiger": molsetup.charge[atom.index],
        }
    return data


def validate_key(key, keyword, terms):
    """validate Proper torsion from OFFXML"""

    if not key.startswith(keyword):
        return False
    length = len(keyword)
    if not key[length:].isdigit():  # e.g. key="phase1", keyword="phase", length=5
        return False
    series_index = int(key[length:])
    terms.setdefault(series_index, set())
    terms[series_index].add(keyword)
    return True


def smirks_to_smarts(smirks):
    """returns:
    - SMARTS - same as smirks, but without labels, i.e. [C:1] -> [C]
    - labels - mapping of labels to atom indices, i.e. {1: 0}

    not needed for rdkit, but for openbabel which lacks support for labels
    """

    n_open = 0
    n_close = 0
    depth_recursion = 0  # recursive smarts, e.g. [C$([N])]
    bracket_depth = 0
    last_recursion_bracket_depth = [None]  # bracket depth when '$(' was found
    labels = (
        {}
    )  # , keys: integer label in smirks, values: index of atom in smarts molecule
    chars_to_delete = []
    for i, char in enumerate(smirks):
        if char == "[" and depth_recursion == 0:
            n_open += 1
        elif char == "]" and depth_recursion == 0:
            n_close += 1
            string = ""  # to store chars between : and ]
            candidate_chars_to_delete = []
            for j in range(i):
                k = i - j - 1  # chars up to 'i', in reverse
                c = smirks[k]
                candidate_chars_to_delete.append(k)  # includes ":"
                if c == ":":
                    break
                string = c + string
            if string.isdigit():
                label_id = int(string)
                if label_id in labels:
                    msg = "\ncan't convert smirks with repeated label: %d" % label_id
                    msg += "\nsmirks: %s" % smirks
                    raise ValueError(msg)
                labels[label_id] = n_open - 1
                chars_to_delete.extend(candidate_chars_to_delete)
        elif char == "$" and smirks[i + 1] == "(":
            last_recursion_bracket_depth.append(bracket_depth + 1)
            depth_recursion += 1
        elif char == "(":
            bracket_depth += 1
        elif (
            char == ")"
            and depth_recursion > 0
            and last_recursion_bracket_depth[-1] == bracket_depth
        ):
            depth_recursion -= 1
            last_recursion_bracket_depth.pop(-1)
            bracket_depth -= 1
        elif char == ")":
            bracket_depth -= 1

    if n_open != n_close:
        raise RuntimeError("ooops different number of [ and ]")
    mol = Chem.MolFromSmarts(smirks)
    n_atoms = mol.GetNumAtoms()
    if n_open != n_atoms:
        print(n_open, n_atoms)
        raise RuntimeError("need all atoms to be encapsulated by [ ]")

    assert len(set(chars_to_delete)) == len(chars_to_delete)

    smarts = list(smirks)
    for i in sorted(chars_to_delete)[::-1]:
        smarts.pop(i)
    smarts = "".join(smarts)
    return smarts, labels


def make_dihedral_entry(attrib_dict_from_xml):
    """convert 'Proper' dict from OFFXML to autodockdev dict"""

    smirks = attrib_dict_from_xml["smirks"]
    smarts, labels = smirks_to_smarts(smirks)
    assert len(labels) == 4
    assert 1 in labels and 2 in labels and 3 in labels and 4 in labels
    dihedral_entry = {
        "smarts": smarts,
        "IDX": [labels[index] for index in (1, 2, 3, 4)],
    }
    terms = {}  # keywords found for each term of the fourier series
    expected_keywords = set(["k", "phase", "periodicity", "idivf"])
    for key, value in attrib_dict_from_xml.items():
        if key == "smirks":
            continue
        elif key == "id":
            dihedral_entry["id"] = value
        else:
            for keyword in expected_keywords:
                is_valid_key = validate_key(key, keyword, terms)
                if is_valid_key:
                    break
            if not is_valid_key:
                msg = "\nGot unexpected key: %s" % key
                msg += "\nOffending 'Proper' entry: %s" % attrib_dict_from_xml
                raise ValueError(msg)
            elif keyword == "k":
                dihedral_entry[key] = float(
                    value.replace("* mole**-1 * kilocalorie", "")
                )
            elif keyword == "phase":
                dihedral_entry[key] = math.radians(float(value.replace("* degree", "")))
            elif keyword == "periodicity":
                dihedral_entry[key] = int(value)
            elif keyword == "idivf":
                dihedral_entry[key] = float(value)
    # check that all terms in the fourier series have all expected keywords
    assert len(terms) >= 1
    for series_index in terms:
        if terms[series_index] != expected_keywords:
            msg = "\nmismatch between input keywords: %s\n" % terms[series_index]
            msg += "           and expected keywords: %s\n" % expected_keywords
            msg += "               with series_index: %d" % series_index
            raise ValueError(msg)
    return dihedral_entry


def make_vdw_entry(attrib_dict_from_xml):
    """convert 'Atom' dict from OFFXML to autodockdev dict"""

    smirks = attrib_dict_from_xml["smirks"]
    smarts, labels = smirks_to_smarts(smirks)
    assert len(labels) == 1
    assert 1 in labels
    vdw_entry = {
        "smarts": smarts,
        "IDX": [labels[1]],
        "id": attrib_dict_from_xml["id"],
    }
    assert ("epsilon" in attrib_dict_from_xml) and (
        ("rmin_half" in attrib_dict_from_xml) ^ ("sigma" in attrib_dict_from_xml)
    )
    for key, value in attrib_dict_from_xml.items():
        if key in ["id", "smirks"]:
            continue
        elif key == "epsilon":
            vdw_entry["epsilon"] = float(value.replace("* mole**-1 * kilocalorie", ""))
        elif key == "rmin_half":
            vdw_entry["rmin_half"] = float(value.replace("* angstrom", ""))
        elif key == "sigma":
            vdw_entry["rmin_half"] = (
                float(value.replace("* angstrom", "")) * (4 ** (1.0 / 12)) * 0.5
            )
        else:
            msg = "\nGot unexpected key: %s" % key
            msg += "\nOffending 'Atom' entry: %s" % attrib_dict_from_xml
            raise ValueError(msg)
    return vdw_entry


def assign_atypes(vdw_list, use_openff_id=True, force_uppercase=True):

    atypes_preserve = ["n-tip3p-O", "n-tip3p-H"]

    number_by_element = {}
    used_numbers = {}
    atomic_numbers = []

    for v in vdw_list:
        mol = Chem.MolFromSmarts(v["smarts"])
        atom = mol.GetAtomWithIdx(
            v["IDX"][0]
        )  # consider only the first if multiple IDX
        element = mini_periodic_table[atom.GetAtomicNum()]
        atomic_numbers.append(atom.GetAtomicNum())
        used_numbers.setdefault(element, set())
        off_id = v["id"]

        if off_id in atypes_preserve:
            atype = off_id
        elif use_openff_id:
            # use id ("n1", "n2", "n3") -> [H1, H2, C3]
            if off_id[0] == "n" and off_id[1:].isdigit():
                n = int(off_id[1:])
            else:
                n = len(vdw_list) + 1
            while n in used_numbers[element]:
                n += 1
            used_numbers[element].add(n)
            atype = "%s%d" % (element, n)
        else:
            # each element starts from 1 -> [H1, H2, C1]
            number_by_element.setdefault(element, 0)
            number_by_element[element] += 1
            n = number_by_element[element]
            atype = "%s%d" % (element, n)

        if force_uppercase:
            v["atype"] = atype.upper()
        else:
            v["atype"] = atype
    return atomic_numbers  # needed to get atomic mass aftwerwards


def make_vdw_by_atype(vdw_list, atomic_numbers):
    bytype = {}
    index = 0
    for v in vdw_list:
        atype = v["atype"]
        param = {
            "rmin_half": v["rmin_half"],
            "epsilon": v["epsilon"],
            "atomic_number": atomic_numbers[index],
        }
        if atype in bytype:
            assert bytype[atype] == param
        else:
            bytype[atype] = param
        index += 1
    return bytype


def parse_offxml(offxml_filename):
    """
    Convert OpenFF XML entries to autodockdev dictionaries
    """

    root = ET.parse(offxml_filename).getroot()

    torsions = root.findall("ProperTorsions")
    vdw = root.findall("vdW")
    assert len(torsions) == 1
    assert len(vdw) == 1
    torsions = torsions[0]
    vdw = vdw[0]

    vdw_list = []
    for child in vdw:
        if child.tag != "Atom":
            print(" SKIPPING: %s" % child.tag)
            continue
        v = make_vdw_entry(child.attrib)
        vdw_list.append(v)
    atomic_numbers = assign_atypes(vdw_list)
    vdw_by_atype = make_vdw_by_atype(vdw_list, atomic_numbers)

    dihedral_list = []
    for child in torsions:
        if child.tag != "Proper":
            print(" SKIPPING: %s" % child.tag)
            continue
        d = make_dihedral_entry(child.attrib)
        dihedral_list.append(d)

    return vdw_list, dihedral_list, vdw_by_atype



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

__version__ = "0.6.1"

try:
    import prody
except ImportError:
    _has_prody = False
else:
    _has_prody = True

from .preparation import MoleculePreparation
from .molsetup import RDKitMoleculeSetup
from .molsetup import MoleculeSetup
from .molsetup import MoleculeSetupEncoder
from .molsetup import Restraint
from .molsetup import UniqAtomParams
from .utils import rdkitutils
from .utils import pdbutils
from .utils import geomutils
from .utils import utils
from .atomtyper import AtomTyper
from .receptor_pdbqt import PDBQTReceptor
from .polymer import Polymer
from .polymer import Monomer
from .polymer import ResiduePadder
from .polymer import ResidueTemplate
from .polymer import ResidueChemTemplates
from .polymer import PolymerEncoder
from .polymer import MonomerEncoder
from .polymer import ResiduePadderEncoder
from .polymer import ResidueTemplateEncoder
from .polymer import ResidueChemTemplatesEncoder
from .polymer import add_rotamers_to_polymer_molsetups
from .polymer import polymer_json_decoder
from .polymer import PolymerCreationError
from .molecule_pdbqt import PDBQTMolecule
from .rdkit_mol_create import RDKitMolCreate
from .export_flexres import export_pdb_updated_flexres
from .reactive import reactive_typer
from .reactive import get_reactive_config
from .writer import PDBQTWriterLegacy
from . import analysis
from .writer import oids_block_from_setup
from .openff_xml_parser import parse_offxml
from .openff_xml_parser import load_openff
from .openff_xml_parser import get_openff_epsilon_sigma
from .hydrate import Hydrate

import logging
from rdkit import rdBase
rdkit_logger = logging.getLogger("rdkit")
rdkit_logger.handlers[0].setLevel("WARNING")
rdkit_logger.handlers[0].setFormatter(
    logging.Formatter('[RDKit] %(levelname)s:%(message)s'),
)
rdBase.LogToPythonLogger()

__all__ = ['MoleculePreparation', 'RDKitMoleculeSetup', 'MoleculeSetupEncoder',
           'pdbutils', 'geomutils', 'rdkitutils', 'utils',
           'AtomTyper', 'PDBQTMolecule', 'PDBQTReceptor', 'analysis',
           'Polymer', 'Monomer', 'ResiduePadder', 'ResidueTemplate', 'ResidueChemTemplates',
           'PolymerEncoder', 'MonomerEncoder', 'ResiduePadderEncoder', 'ResidueTemplateEncoder',
           'ResidueChemTemplatesEncoder',
           'add_rotamers_to_polymer_molsetups',
           'RDKitMolCreate',
           'PDBQTWriterLegacy',
           'reactive_typer',
           'get_reactive_config',
           'gridbox',
           'oids_block_from_setup',
           'parse_offxml',
           'Hydrate',
           'Restraint',
           ]

if _has_prody:
    from .covalentbuilder import CovalentBuilder
    __all__.append("CovalentBuilder")



import collections
import json
import meeko
import numpy
import pathlib
import pytest

from meeko import (
    Monomer,
    MonomerEncoder,
    Polymer,
    PolymerEncoder,
    MoleculePreparation,
    MoleculeSetup,
    MoleculeSetupEncoder,
    RDKitMoleculeSetup,
    ResiduePadder,
    ResiduePadderEncoder,
    ResidueTemplate,
    ResidueTemplateEncoder,
    ResidueChemTemplates,
    ResidueChemTemplatesEncoder,
    PDBQTWriterLegacy,
)

from meeko import polymer
from meeko.molsetup import Atom, Bond, Ring, RingClosureInfo, Restraint

from rdkit import Chem
from rdkit.Chem import rdChemReactions

from meeko.utils.pdbutils import PDBAtomInfo

# from ..meeko.utils.pdbutils import PDBAtomInfo

pkgdir = pathlib.Path(meeko.__file__).parents[1]

# Test Data
ahhy_example = pkgdir / "test/polymer_data/AHHY.pdb"
just_one_ALA_missing = (
    pkgdir / "test/polymer_data/just-one-ALA-missing-CB.pdb"
)

# Polymer creation data
chem_templates = ResidueChemTemplates.create_from_defaults()
mk_prep = MoleculePreparation()


# region Fixtures
@pytest.fixture
def populated_polymer():
    file = open(ahhy_example)
    pdb_str = file.read()
    polymer = Polymer.from_pdb_string(
        pdb_str, chem_templates, mk_prep, blunt_ends=[("A:1", 0)]
    )
    return polymer


@pytest.fixture
def populated_polymer_missing():
    file = open(just_one_ALA_missing)
    pdb_str = file.read()
    polymer = Polymer.from_pdb_string(
        pdb_str,
        chem_templates,
        mk_prep,
        blunt_ends=[("A:1", 0), ("A:1", 2)],
        allow_bad_res=True,
    )
    return polymer


@pytest.fixture
def populated_monomer(populated_polymer):
    polymer = populated_polymer
    return polymer.monomers["A:1"]


@pytest.fixture
def populated_rdkit_molsetup(populated_monomer):
    monomer = populated_monomer
    return monomer.molsetup


@pytest.fixture
def populated_residue_chem_templates(populated_polymer):
    polymer = populated_polymer
    return polymer.residue_chem_templates


@pytest.fixture
def populated_residue_template(populated_residue_chem_templates):
    res_chem_templates = populated_residue_chem_templates
    return res_chem_templates.residue_templates["G"]


@pytest.fixture
def populated_residue_padder(populated_residue_chem_templates):
    res_chem_templates = populated_residue_chem_templates
    return res_chem_templates.padders["5-prime"]


# endregion


# region Test Cases
def test_rdkit_molsetup_encoding_decoding(populated_rdkit_molsetup):
    """
    Takes a fully populated RDKitMoleculeSetup, checks that it can be serialized to JSON and deserialized back into an
    object without any errors, then checks that the deserialized object matches the starting object and that the
    attribute types, values, and structure of the deserialized object are as expected for an RDKitMoleculeSetup.

    Parameters
    ----------
    populated_rdkit_molsetup: RDKitMoleculeSetup
        Takes as input a populated RDKitMoleculeSetup object.

    Returns
    -------
    None
    """
    # TODO: Certain fields are empty in this example, and if we want to make sure that json is working in all scenarios
    # we will need to make other tests for those empty fields.
    # Encode and decode MoleculeSetup from json
    starting_molsetup = populated_rdkit_molsetup
    json_str = json.dumps(starting_molsetup, cls=MoleculeSetupEncoder)
    decoded_molsetup = json.loads(json_str, object_hook=RDKitMoleculeSetup.from_json)

    # First asserts that all types are as expected
    assert isinstance(starting_molsetup, RDKitMoleculeSetup)
    assert isinstance(decoded_molsetup, RDKitMoleculeSetup)

    # Go through MoleculeSetup attributes and check that they are the expected type and match the MoleculeSetup object
    # before serialization.
    check_molsetup_equality(decoded_molsetup, starting_molsetup)
    return


def test_monomer_encoding_decoding(populated_monomer):
    """
    Takes a fully populated Monomer, checks that it can be serialized to JSON and deserialized back into an
    object without any errors, then checks that the deserialized object matches the starting object and that the
    attribute types, values, and structure of the deserialized object are as expected for an Monomer.

    Parameters
    ----------
    populated_monomer: Monomer
        Takes as input a populated Monomer object.

    Returns
    -------
    None
    """
    # Starts by getting a Monomer object, converting it to a json string, and then decoding the string into
    # a new Monomer object
    starting_monomer = populated_monomer
    json_str = json.dumps(starting_monomer, cls=MonomerEncoder)
    decoded_monomer = json.loads(
        json_str, object_hook=polymer.monomer_json_decoder
    )

    # Asserts that the starting and ending objects have the expected Monomer type
    assert isinstance(starting_monomer, Monomer)
    assert isinstance(decoded_monomer, Monomer)

    check_monomer_equality(decoded_monomer, starting_monomer)
    return


def test_pdbqt_writing_from_decoded_polymer(populated_polymer):
    """
    Takes a fully populated Polymer, writes a PDBQT string from it, encodes and decodes it, writes
    another PDBQT string from the decoded polymer, and then checks that the PDBQT strings are identical.

    Parameters
    ----------
    populated_polymer: Polymer
        Takes as input a populated Polymer object.

    Returns
    -------
    None
    """

    starting_polymer = populated_polymer
    starting_pdbqt = PDBQTWriterLegacy.write_from_polymer(starting_polymer)
    json_str = json.dumps(starting_polymer, cls=PolymerEncoder)
    decoded_polymer = json.loads(
        json_str, object_hook=polymer.polymer_json_decoder
    )
    decoded_pdbqt = PDBQTWriterLegacy.write_from_polymer(decoded_polymer) 
    assert decoded_pdbqt == starting_pdbqt
    return



def test_residue_template_encoding_decoding(populated_residue_template):
    """
    Takes a fully populated ResidueTemplate, checks that it can be serialized to JSON and deserialized back into an
    object without any errors, then checks that the deserialized object matches the starting object and that the
    attribute types, values, and structure of the deserialized object are as expected for an ResidueTemplate.

    Parameters
    ----------
    populated_residue_template: ResidueTemplate
        Takes as input a populated ResidueTemplate object.

    Returns
    -------
    None
    """
    # Starts by getting a ResidueTemplate object, converting it to a json string, and then decoding the string into
    # a new ResidueTemplate object
    starting_template = populated_residue_template
    json_str = json.dumps(starting_template, cls=ResidueTemplateEncoder)
    decoded_template = json.loads(
        json_str, object_hook=polymer.residue_template_json_decoder
    )

    # Asserts that the starting and ending objects have the expected ResidueTemplate type
    assert isinstance(starting_template, ResidueTemplate)
    assert isinstance(decoded_template, ResidueTemplate)

    # Checks that the two residue templates are equal
    check_residue_template_equality(decoded_template, starting_template)
    return


def test_residue_padder_encoding_decoding(populated_residue_padder):
    """
    Takes a fully populated ResiduePadder, checks that it can be serialized to JSON and deserialized back into an
    object without any errors, then checks that the deserialized object matches the starting object and that the
    attribute types, values, and structure of the deserialized object are as expected for an ResiduePadder.

    Parameters
    ----------
    populated_residue_padder: ResiduePadder
        Takes as input a populated ResiduePadder object.

    Returns
    -------
    None
    """
    # Starts by getting a ResiduePadder object, converting it to a json string, and then decoding the string into
    # a new ResiduePadder object
    starting_padder = populated_residue_padder
    json_str = json.dumps(starting_padder, cls=ResiduePadderEncoder)
    decoded_padder = json.loads(
        json_str, object_hook=polymer.residue_padder_json_decoder
    )

    # Asserts that the starting and ending objects have the expected ResiduePadder type
    assert isinstance(starting_padder, ResiduePadder)
    assert isinstance(decoded_padder, ResiduePadder)

    # Checks that the two residue padders are equal
    check_residue_padder_equality(decoded_padder, starting_padder)
    return


def test_residue_chem_templates_encoding_decoding(populated_residue_chem_templates):
    """
    Takes a fully populated ResidueChemTemplates, checks that it can be serialized to JSON and deserialized back into an
    object without any errors, then checks that the deserialized object matches the starting object and that the
    attribute types, values, and structure of the deserialized object are as expected for an ResidueChemTemplates.

    Parameters
    ----------
    populated_residue_chem_templates: ResidueChemTemplates
        Takes as input a populated ResidueChemTemplates object.

    Returns
    -------
    None
    """
    # Starts by getting a ResidueChemTemplates object, converting it to a json string, and then decoding the string into
    # a new ResidueChemTemplates object
    starting_templates = populated_residue_chem_templates
    json_str = json.dumps(starting_templates, cls=ResidueChemTemplatesEncoder)
    decoded_templates = json.loads(
        json_str, object_hook=polymer.residue_chem_templates_json_decoder
    )

    # Asserts that the starting and ending objects have the expected ResidueChemTemplates type
    assert isinstance(starting_templates, ResidueChemTemplates)
    assert isinstance(decoded_templates, ResidueChemTemplates)

    # Checks that the two chem templates are equal
    check_residue_chem_templates_equality(decoded_templates, starting_templates)
    return


def test_polymer_encoding_decoding(
    populated_polymer, populated_polymer_missing
):
    """
    Takes a fully populated Polymer, checks that it can be serialized to JSON and deserialized back into an
    object without any errors, then checks that the deserialized object matches the starting object and that the
    attribute types, values, and structure of the deserialized object are as expected for a Polymer.

    Parameters
    ----------
    populated_polymer: Polymer
        Takes as input a populated Polymer object.

    Returns
    -------
    None
    """
    # Starts by getting a Polymer object, converting it to a json string, and then decoding the string into
    # a new Polymer object
    polymers = (
        populated_polymer,
        populated_polymer_missing,
    )
    for starting_polymer in polymers:
        json_str = json.dumps(starting_polymer, cls=PolymerEncoder)
        decoded_polymer = json.loads(
            json_str, object_hook=polymer.polymer_json_decoder
        )

        # Asserts that the starting and ending objects have the expected Polymer type
        assert isinstance(starting_polymer, Polymer)
        assert isinstance(decoded_polymer, Polymer)

        # Checks that the two polymers are equal
        check_polymer_equality(decoded_polymer, starting_polymer)
    return


# endregion


# region Object Equality Checks
def check_molsetup_equality(decoded_obj: MoleculeSetup, starting_obj: MoleculeSetup):
    """
    Asserts that two MoleculeSetup objects are equal, and that the decoded_obj input has fields contain correctly typed
    data.

    Parameters
    ----------
    decoded_obj: MoleculeSetup
        A MoleculeSetup object that we want to check is correctly typed and contains the correct data.
    starting_obj: MoleculeSetup
        A MoleculeSetup object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """

    # Checks if the MoleculeSetup is an RDKitMoleculeSetup, and if so also checks the RDKitMoleculeSetup attributes
    if isinstance(starting_obj, RDKitMoleculeSetup):
        assert isinstance(decoded_obj.mol, Chem.rdchem.Mol)
        pass

    # Going through and checking MoleculeSetup attributes
    assert decoded_obj.name == starting_obj.name
    assert isinstance(decoded_obj.is_sidechain, bool)
    assert decoded_obj.is_sidechain == starting_obj.is_sidechain
    assert isinstance(decoded_obj.pseudoatom_count, int)
    assert decoded_obj.pseudoatom_count == starting_obj.pseudoatom_count

    # Checking atoms
    atom_idx = 0
    assert len(decoded_obj.atoms) == len(starting_obj.atoms)
    for atom in decoded_obj.atoms:
        assert isinstance(atom, Atom)
        assert atom.index == atom_idx
        check_atom_equality(atom, starting_obj.atoms[atom_idx])
        atom_idx += 1

    # Checking bonds
    for bond_id in starting_obj.bond_info:
        assert isinstance(decoded_obj.bond_info[bond_id], Bond)
        assert bond_id in decoded_obj.bond_info
        check_bond_equality(
            decoded_obj.bond_info[bond_id], starting_obj.bond_info[bond_id]
        )

    # Checking rings
    for ring_id in starting_obj.rings:
        assert isinstance(decoded_obj.rings[ring_id], Ring)
        assert ring_id in decoded_obj.rings
        check_ring_equality(decoded_obj.rings[ring_id], starting_obj.rings[ring_id])
    assert isinstance(decoded_obj.ring_closure_info, RingClosureInfo)
    assert (
        decoded_obj.ring_closure_info.bonds_removed
        == starting_obj.ring_closure_info.bonds_removed
    )
    for key in starting_obj.ring_closure_info.pseudos_by_atom:
        assert key in decoded_obj.ring_closure_info.pseudos_by_atom
        assert (
            decoded_obj.ring_closure_info.pseudos_by_atom[key]
            == starting_obj.ring_closure_info.pseudos_by_atom[key]
        )

    # Checking other fields
    assert len(decoded_obj.rotamers) == len(starting_obj.rotamers)
    for idx, component_dict in enumerate(starting_obj.rotamers):
        decoded_dict = decoded_obj.rotamers[idx]
        for key in component_dict:
            assert key in decoded_dict
            assert decoded_dict[key] == component_dict[key]
    for key in starting_obj.atom_params:
        assert key in decoded_obj.atom_params
        assert decoded_obj.atom_params[key] == starting_obj.atom_params[key]
    assert len(decoded_obj.restraints) == len(starting_obj.restraints)
    for idx, restraint in starting_obj.restraints:
        assert isinstance(decoded_obj.restraints[idx], Restraint)
        check_restraint_equality(
            decoded_obj.restraints[idx], starting_obj.restraints[idx]
        )

    # Checking flexibility model
    for key in starting_obj.flexibility_model:
        assert key in decoded_obj.flexibility_model
        assert decoded_obj.flexibility_model[key] == starting_obj.flexibility_model[key]
    return


def check_atom_equality(decoded_obj: Atom, starting_obj: Atom):
    """
    Asserts that two Atom objects are equal, and that the decoded_obj input has fields contain correctly typed
    data.

    Parameters
    ----------
    decoded_obj: Atom
        An Atom object that we want to check is correctly typed and contains the correct data.
    starting_obj: Atom
        An Atom object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    correct_val_type = True
    # np.array conversion checks
    assert isinstance(decoded_obj.coord, numpy.ndarray)
    for i_vec in decoded_obj.interaction_vectors:
        correct_val_type = correct_val_type and isinstance(i_vec, numpy.ndarray)
    assert correct_val_type

    # Checks for equality between decoded and original fields
    assert isinstance(decoded_obj.index, int)
    assert decoded_obj.index == starting_obj.index
    # Only checks pdb info if the starting object's pdbinfo was a string. Otherwise, the decoder is not going to convert
    # the pdbinfo field back to the PDBInfo type right now.
    if isinstance(starting_obj.pdbinfo, str):
        assert decoded_obj.pdbinfo == starting_obj.pdbinfo
    assert isinstance(decoded_obj.charge, float)
    assert decoded_obj.charge == starting_obj.charge
    for idx, val in enumerate(decoded_obj.coord):
        assert val == starting_obj.coord[idx]
    assert isinstance(decoded_obj.atomic_num, int)
    assert decoded_obj.atomic_num == starting_obj.atomic_num
    assert decoded_obj.atom_type == starting_obj.atom_type
    assert decoded_obj.graph == starting_obj.graph
    assert isinstance(decoded_obj.is_ignore, bool)
    assert decoded_obj.is_ignore == starting_obj.is_ignore
    assert isinstance(decoded_obj.is_dummy, bool)
    assert decoded_obj.is_dummy == starting_obj.is_dummy
    assert isinstance(decoded_obj.is_pseudo_atom, bool)
    assert decoded_obj.is_pseudo_atom == starting_obj.is_pseudo_atom
    return


def check_bond_equality(decoded_obj: Bond, starting_obj: Bond):
    """
    Asserts that two Bond objects are equal, and that the decoded_obj input has fields contain correctly typed
    data.

    Parameters
    ----------
    decoded_obj: Bond
        An Bond object that we want to check is correctly typed and contains the correct data.
    starting_obj: Bond
        An Bond object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    assert isinstance(decoded_obj.canon_id, tuple)
    assert isinstance(decoded_obj.canon_id[0], int)
    assert isinstance(decoded_obj.canon_id[1], int)
    assert decoded_obj.canon_id == starting_obj.canon_id
    assert isinstance(decoded_obj.index1, int)
    assert decoded_obj.index1 == starting_obj.index1
    assert isinstance(decoded_obj.index2, int)
    assert decoded_obj.index2 == starting_obj.index2
    assert isinstance(decoded_obj.rotatable, bool)
    assert decoded_obj.rotatable == starting_obj.rotatable
    return


def check_ring_equality(decoded_obj: Ring, starting_obj: Ring):
    """
    Asserts that two Ring objects are equal, and that the decoded_obj input has fields contain correctly typed
    data.

    Parameters
    ----------
    decoded_obj: Ring
        An Ring object that we want to check is correctly typed and contains the correct data.
    starting_obj: Ring
        An Ring object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    assert isinstance(decoded_obj.ring_id, tuple)
    assert decoded_obj.ring_id == starting_obj.ring_id
    assert isinstance(decoded_obj.corner_flip, bool)
    assert decoded_obj.corner_flip == starting_obj.corner_flip
    assert len(decoded_obj.graph) == len(starting_obj.graph)
    for idx, val in enumerate(starting_obj.graph):
        assert decoded_obj.graph[idx] == val
    assert isinstance(decoded_obj.is_aromatic, bool)
    assert decoded_obj.is_aromatic == starting_obj.is_aromatic
    return


def check_restraint_equality(decoded_obj: Restraint, starting_obj: Restraint):
    """
    Asserts that two Restraint objects are equal, and that the decoded_obj input has fields contain correctly typed
    data.

    Parameters
    ----------
    decoded_obj: Restraint
        An Restraint object that we want to check is correctly typed and contains the correct data.
    starting_obj: Restraint
        An Restraint object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    assert isinstance(decoded_obj.atom_index, int)
    assert decoded_obj.atom_index == starting_obj.atom_index
    assert isinstance(decoded_obj.target_coords, tuple)
    assert decoded_obj.target_coords == starting_obj.target_coords
    assert isinstance(decoded_obj.kcal_per_angstrom_square, float)
    assert decoded_obj.kcal_per_angstrom_square == starting_obj.kcal_per_angstrom_square
    assert isinstance(decoded_obj.delay_angstroms, float)
    assert decoded_obj.delay_angstroms == starting_obj.delay_angstroms
    return


def check_monomer_equality(decoded_obj: Monomer, starting_obj: Monomer):
    """
    Asserts that two Monomer objects are equal, and that the decoded_obj input has fields contain correctly typed
    data.

    Parameters
    ----------
    decoded_obj: Monomer
        A Monomer object that we want to check is correctly typed and contains the correct data.
    starting_obj: Monomer
        A Monomer object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    # Goes through the Monomer's fields and checks that they are the expected type and match the Monomer
    # object before serialization (that we have effectively rebuilt the Monomer)

    # RDKit Mols - Check whether we can test for equality with RDKit Mols
    # assert decoded_monomer.raw_rdkit_mol == starting_residue.raw_rdkit_mol
    assert type(decoded_obj.raw_rdkit_mol) == type(starting_obj.raw_rdkit_mol)
    if isinstance(decoded_obj.raw_rdkit_mol, Chem.rdchem.Mol):
        assert Chem.MolToSmiles(decoded_obj.raw_rdkit_mol) == Chem.MolToSmiles(
            starting_obj.raw_rdkit_mol
        )
    # assert decoded_monomer.rdkit_mol == starting_monomer.rdkit_mol
    assert type(decoded_obj.rdkit_mol) == type(starting_obj.rdkit_mol)
    if isinstance(decoded_obj.rdkit_mol, Chem.rdchem.Mol):
        assert Chem.MolToSmiles(decoded_obj.rdkit_mol) == Chem.MolToSmiles(
            starting_obj.rdkit_mol
        )
    # assert decoded_monomer.padded_mol == starting_monomer.padded_mol
    assert type(decoded_obj.padded_mol) == type(starting_obj.padded_mol)
    if isinstance(decoded_obj.padded_mol, Chem.rdchem.Mol):
        assert Chem.MolToSmiles(decoded_obj.padded_mol) == Chem.MolToSmiles(
            starting_obj.padded_mol
        )

    # MapIDX
    assert decoded_obj.mapidx_to_raw == starting_obj.mapidx_to_raw
    assert decoded_obj.mapidx_from_raw == starting_obj.mapidx_from_raw

    # Non-Bool vars
    assert decoded_obj.residue_template_key == starting_obj.residue_template_key
    assert decoded_obj.input_resname == starting_obj.input_resname
    assert decoded_obj.atom_names == starting_obj.atom_names
    assert type(decoded_obj.molsetup) == type(starting_obj.molsetup)
    if isinstance(decoded_obj.molsetup, RDKitMoleculeSetup):
        check_molsetup_equality(decoded_obj.molsetup, starting_obj.molsetup)

    # Bools
    assert decoded_obj.is_flexres_atom == starting_obj.is_flexres_atom
    assert decoded_obj.is_movable == starting_obj.is_movable
    return


def check_residue_chem_templates_equality(
    decoded_obj: ResidueChemTemplates, starting_obj: ResidueChemTemplates
):
    """
    Asserts that two ResidueChemTemplates objects are equal, and that the decoded_obj input has fields contain correctly
    typed data.

    Parameters
    ----------
    decoded_obj: ResidueChemTemplates
        A ResidueChemTemplates object that we want to check is correctly typed and contains the correct data.
    starting_obj: ResidueChemTemplates
        A ResidueChemTemplates object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    # correct_val_type is used to check that all type conversions for nested data have happened correctly
    correct_val_type = True
    # Checks residue_templates by ensuring it has the same members as the starting object, that each value in the
    # dictionary is a ResidueTemplate object, and that each template is equal to its corresponding ResidueTemplate in
    # the starting object.
    assert decoded_obj.residue_templates.keys() == starting_obj.residue_templates.keys()
    for key in decoded_obj.residue_templates:
        correct_val_type = correct_val_type & isinstance(
            decoded_obj.residue_templates[key], ResidueTemplate
        )
        check_residue_template_equality(
            decoded_obj.residue_templates[key], starting_obj.residue_templates[key]
        )
    assert correct_val_type

    # Directly compares ambiguous values.
    assert decoded_obj.ambiguous == starting_obj.ambiguous

    # Checks padders by ensuring it has the same members as the starting object, that each value in the dictionary is a
    # ResiduePadder object, and that each padder is equal to its corresponding ResiduePadder in the starting object.
    assert decoded_obj.padders.keys() == starting_obj.padders.keys()
    for key in decoded_obj.padders:
        correct_val_type = correct_val_type & isinstance(
            decoded_obj.padders[key], ResiduePadder
        )
        check_residue_padder_equality(
            decoded_obj.padders[key], starting_obj.padders[key]
        )
    assert correct_val_type
    return


def check_residue_template_equality(
    decoded_obj: ResidueTemplate, starting_obj: ResidueTemplate
):
    """
    Asserts that two ResidueTemplate objects are equal, and that the decoded_obj input has fields contain correctly typed
    data.

    Parameters
    ----------
    decoded_obj: ResidueTemplate
        A ResidueTemplate object that we want to check is correctly typed and contains the correct data.
    starting_obj: ResidueTemplate
        A ResidueTemplate object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    # Goes through the ResidueTemplate's fields and checks that they have the expected type and that they match the
    # ResidueTemplate object before serialization
    assert isinstance(decoded_obj.mol, Chem.rdchem.Mol)

    assert decoded_obj.link_labels == starting_obj.link_labels
    assert decoded_obj.atom_names == starting_obj.atom_names
    return


def check_residue_padder_equality(
    decoded_obj: ResiduePadder, starting_obj: ResiduePadder
):
    """
    Asserts that two ResiduePadder objects are equal, and that the decoded_obj input has fields contain correctly typed
    data.

    Parameters
    ----------
    decoded_obj: ResiduePadder
        A ResiduePadder object that we want to check is correctly typed and contains the correct data.
    starting_obj: ResiduePadder
        A ResiduePadder object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    assert isinstance(decoded_obj.rxn, rdChemReactions.ChemicalReaction)
    decoded_obj_rxn_smarts = rdChemReactions.ReactionToSmarts(decoded_obj.rxn)
    starting_obj_rxn_smarts = rdChemReactions.ReactionToSmarts(starting_obj.rxn)
    assert decoded_obj_rxn_smarts == starting_obj_rxn_smarts

    assert (
        decoded_obj.adjacent_smartsmol_mapidx == starting_obj.adjacent_smartsmol_mapidx
    )

    decoded_adj = decoded_obj.adjacent_smartsmol
    starting_adj = starting_obj.adjacent_smartsmol
    assert isinstance(decoded_adj, Chem.rdchem.Mol) or decoded_adj is None
    if decoded_adj is None:
        assert decoded_adj == starting_adj
    else:
        decoded_adj_smarts = Chem.MolToSmarts(decoded_adj)
        starting_adj_smarts = Chem.MolToSmarts(starting_adj)
        assert decoded_adj_smarts == starting_adj_smarts
    return


def check_polymer_equality(
    decoded_obj: Polymer, starting_obj: Polymer
):
    """
    Asserts that two Polymer objects are equal, and that the decoded_obj input has fields contain correctly
    typed data.

    Parameters
    ----------
    decoded_obj: Polymer
        A Polymer object that we want to check is correctly typed and contains the correct data.
    starting_obj: Polymer
        A Polymer object with the desired values to check the decoded object against.

    Returns
    -------
    None
    """
    # correct_val_type is used to check that all type conversions for nested data have happened correctly
    correct_val_type = True
    # Checks residue_chem_templates equality
    check_residue_chem_templates_equality(
        decoded_obj.residue_chem_templates, starting_obj.residue_chem_templates
    )

    # Loops through residues, checks that the decoded and starting obj share the same set of keys, that all the residues
    # are represented as Monomer objects, and that the decoding and starting obj Monomers are equal.
    assert decoded_obj.monomers.keys() == starting_obj.monomers.keys()
    for key in decoded_obj.monomers:
        correct_val_type = correct_val_type & isinstance(
            decoded_obj.monomers[key], Monomer
        )
        check_monomer_equality(decoded_obj.monomers[key], starting_obj.monomers[key])
    assert correct_val_type

    # Checks log equality
    assert decoded_obj.log == starting_obj.log
    return


# endregion



from rdkit import Chem
from rdkit.Chem import rdMolInterchange


SERIALIZATION_SEPARATOR_CHAR = ","


def rdkit_mol_from_json(json_str: str):
    """
    Takes in a JSON string and attempts to use RDKit's JSON to Mols utility to extract just one RDKitMol from the
    json string. If none or more than one Mols are returned, raises an error.

    Parameters
    ----------
    json_str: str
        A JSON string representing an RDKit Mol.

    Returns
    -------
    rdkit_mol: rdkit.Chem.rdchem.Mol
        An RDKit Mol object corresponding to the input JSON string

    Raises
    ------
    ValueError
        If no RDKitMol objects are returned, or if more than one is returned, throws a ValueError.
    """
    if json_str is None:
        return None
    rdkit_mols = rdMolInterchange.JSONToMols(json_str)
    if len(rdkit_mols) != 1:
        raise ValueError(
            f"Expected 1 rdkit mol from json string but got {len(rdkit_mols)}"
        )
    Chem.SanitizeMol(rdkit_mols[0])  # needed to compute gasteiger charges
    return rdkit_mols[0]


def tuple_to_string(input_tuple: tuple):
    """
    Converts a tuple to a JSON serializable string.

    Parameters
    ----------
    input_tuple: tuple
        A tuple to convert to a JSON serializable string.

    Returns
    -------
    A string representation of the tuple using the specified serialization separator character.
    """
    return SERIALIZATION_SEPARATOR_CHAR.join([str(i) for i in input_tuple])


def string_to_tuple(input_string: str, element_type: type = str):
    """
    Takes a JSON string and converts it back to a tuple. If element type is specified, converts all elements of the
    tuple to that type.

    Parameters
    ----------
    input_string: str
        String deserialized from JSON.
    element_type: type
        Data type for all of the elements of the tuple.

    Returns
    -------
    A deserialized tuple with the specified element type.
    """
    if element_type is not str:
        return tuple(
            [element_type(i) for i in input_string.split(SERIALIZATION_SEPARATOR_CHAR)]
        )
    else:
        return tuple(input_string)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko preparation
#

from inspect import signature
import json
from os import linesep as eol
import pathlib
import warnings

from rdkit import Chem

import meeko.macrocycle
from .molsetup import Bond
from .molsetup import RDKitMoleculeSetup
from .atomtyper import AtomTyper
from .espalomatyper import EspalomaTyper
from .bondtyper import BondTyperLegacy
from .hydrate import HydrateMoleculeLegacy
from .macrocycle import FlexMacrocycle
from .flexibility import get_flexibility_model
from .flexibility import update_closure_atoms
from .flexibility import merge_terminal_atoms
from .writer import PDBQTWriterLegacy
from .reactive import assign_reactive_types
from .openff_xml_parser import load_openff

pkg_dir = pathlib.Path(__file__).parents[0]
params_dir = pkg_dir / "data" / "params"
# the above is controversial, see
# https://stackoverflow.com/questions/6028000/how-to-read-a-static-file-from-inside-a-python-package

# DeprecationWarning is not displayed by default
warnings.filterwarnings("default", category=DeprecationWarning)


class MoleculePreparation:
    """
    Attributes
    ----------
    deprecated_setup_access:
    merge_these_atom_types: tuple
    hydrate: bool
    flexible_amides: bool
    rigid_macrocycles: bool
    min_ring_size: int
    max_ring_size: int
    keep_chorded_rings: bool
    keep_equivalent_rings: bool
    double_bond_penalty: float
    macrocycle_allow_A: bool
    rigidify_bonds_smarts: list
    rigidify_bonds_indices: list

    input_atom_params:
    load_atom_params:
    add_atom_types:

    atom_params:
    """

    packaged_params = {}
    for path in params_dir.glob("*.json"):  # e.g. data/params/ad4_types.json
        name = path.with_suffix("").name  # e.g. "ad4_types"
        packaged_params[name] = path

    def __init__(
        self,
        merge_these_atom_types=("H",),
        hydrate=False,
        flexible_amides=False,
        rigid_macrocycles=False,
        min_ring_size=meeko.macrocycle.DEFAULT_MIN_RING_SIZE,
        max_ring_size=meeko.macrocycle.DEFAULT_MAX_RING_SIZE,
        keep_chorded_rings=False,
        keep_equivalent_rings=False,
        double_bond_penalty=meeko.macrocycle.DEFAULT_DOUBLE_BOND_PENALTY,
        macrocycle_allow_A=False,
        rigidify_bonds_smarts=[],
        rigidify_bonds_indices=[],
        input_atom_params=None,
        load_atom_params="ad4_types",
        add_atom_types=(),
        input_offatom_params=None,
        load_offatom_params=None,
        charge_model="gasteiger",
        dihedral_model=None,
        reactive_smarts=None,
        reactive_smarts_idx=None,
        add_index_map=False,
        remove_smiles=False,
    ):
        """

        Parameters
        ----------
        merge_these_atom_types
        hydrate
        flexible_amides
        rigid_macrocycles
        min_ring_size
        max_ring_size
        keep_chorded_rings
        keep_equivalent_rings
        double_bond_penalty
        macrocycle_allow_A
        rigidify_bonds_smarts
        rigidify_bonds_indices
        input_atom_params
        load_atom_params
        add_atom_types
        input_offatom_params
        load_offatom_params
        charge_model
        dihedral_model
        reactive_smarts
        reactive_smarts_idx
        add_index_map
        remove_smiles
        """

        self.deprecated_setup_access = None
        self.merge_these_atom_types = merge_these_atom_types
        self.hydrate = hydrate
        self.flexible_amides = flexible_amides
        self.rigid_macrocycles = rigid_macrocycles
        self.min_ring_size = min_ring_size
        self.max_ring_size = max_ring_size
        self.keep_chorded_rings = keep_chorded_rings
        self.keep_equivalent_rings = keep_equivalent_rings
        self.double_bond_penalty = double_bond_penalty
        self.macrocycle_allow_A = macrocycle_allow_A
        self.rigidify_bonds_smarts = rigidify_bonds_smarts
        self.rigidify_bonds_indices = rigidify_bonds_indices

        self.input_atom_params = input_atom_params
        self.load_atom_params = load_atom_params
        self.add_atom_types = add_atom_types

        self.atom_params = self.get_atom_params(
            input_atom_params, load_atom_params, add_atom_types, self.packaged_params
        )

        if load_offatom_params is not None:
            raise NotImplementedError("load_offatom_params not implemented")
        self.load_offatom_params = load_offatom_params

        allowed_charge_models = ["espaloma", "gasteiger", "zero"]
        if charge_model not in allowed_charge_models:
            raise ValueError(
                "unrecognized charge_model: %s, allowed options are: %s"
                % (charge_model, allowed_charge_models)
            )

        self.charge_model = charge_model

        allowed_dihedral_models = [None, "openff", "espaloma"]
        if dihedral_model in (None, "espaloma"):
            dihedral_list = []
        elif dihedral_model == "openff":
            _, dihedral_list, _ = load_openff()
        else:
            raise ValueError(
                "unrecognized dihedral_model: %s, allowed options are: %s"
                % (dihedral_model, allowed_dihedral_models)
            )

        self.dihedral_model = dihedral_model
        self.dihedral_params = dihedral_list

        if dihedral_model == "espaloma" or charge_model == "espaloma":
            self.espaloma_model = EspalomaTyper()

        self.reactive_smarts = reactive_smarts
        self.reactive_smarts_idx = reactive_smarts_idx
        self.add_index_map = add_index_map
        self.remove_smiles = remove_smiles

        self._bond_typer = BondTyperLegacy()
        self._macrocycle_typer = FlexMacrocycle(
            self.min_ring_size,
            self.max_ring_size,
            self.double_bond_penalty,
            allow_break_atype_A=self.macrocycle_allow_A,
        )
        self._water_builder = HydrateMoleculeLegacy()
        self._classes_setup = {Chem.rdchem.Mol: RDKitMoleculeSetup}

        if input_offatom_params is None:
            self.offatom_params = {}
        else:
            self.offatom_params = input_offatom_params

        if keep_chorded_rings and not keep_equivalent_rings:
            warnings.warn(
                "keep_equivalent_rings=False ignored because keep_chorded_rings=True",
                RuntimeWarning,
            )
        if (reactive_smarts is None) != (reactive_smarts_idx is None):
            raise ValueError(
                "reactive_smarts and reactive_smarts_idx require each other"
            )

    @classmethod
    def from_config(cls, config):
        """

        Parameters
        ----------
        config

        Returns
        -------

        """
        expected_keys = cls.get_defaults_dict().keys()
        bad_keys = [k for k in config if k not in expected_keys]
        if len(bad_keys) > 0:
            err_msg = (
                "unexpected keys in MoleculePreparation.from_config():" + eol
            )
            for key in bad_keys:
                err_msg += "  - %s" % key + eol
            raise ValueError(err_msg)
        p = cls(**config)
        return p

    def calc_flex(
        self,
        setup,
        root_atom_index=None,
        not_terminal_atoms=None,
        delete_ring_bonds=None,
        glue_pseudo_atoms=None,
    ):
        """

        Parameters
        ----------
        setup
        root_atom_index
        not_terminal_atoms
        delete_ring_bonds
        glue_pseudo_atoms

        Returns
        -------

        """
        if not_terminal_atoms is None:
            not_terminal_atoms = []
        if delete_ring_bonds is None:
            delete_ring_bonds = []
        # 5.  break macrocycles into open/linear form
        if self.rigid_macrocycles:
            break_combo_data = None
            bonds_in_rigid_rings = set()
            # every ring is rigid without macrocycle option
            for ring in setup.rings:
                for bond in setup.get_bonds_in_ring(ring):
                    bonds_in_rigid_rings.add(bond)
        else:
            break_combo_data, bonds_in_rigid_rings = (
                self._macrocycle_typer.search_macrocycle(setup, delete_ring_bonds)
            )

        # This must be done before calling get_flexibility_model
        for bond in bonds_in_rigid_rings:
            setup.bond_info[bond].rotatable = False

        flex_model, bonds_to_break = get_flexibility_model(
            setup, root_atom_index, break_combo_data
        )

        # disasble rotatable bonds that rotate nothing (e.g. -CH3 without H)
        # but glue atoms (i.e. CG) are manually marked as non terminal (by
        # passing them in the `glue_atoms` list) to guarantee that the bond
        # to a CG atom is rotatable and the G pseudo rotates
        glue_atoms = []
        for pair in bonds_to_break:
            for index in pair:
                glue_atoms.append(index)
        merge_terminal_atoms(flex_model, not_terminal_atoms + glue_atoms)

        # bond to a terminal atom, or in a ring that isn't flexible
        actual_rotatable = [v for k, v in flex_model["rigid_body_connectivity"].items()]
        actual_rotatable.extend(bonds_to_break)
        for bond in setup.bond_info:
            if bond not in actual_rotatable:
                setup.bond_info[bond].rotatable = False

        # calculate torsions that would be rotatable without macrocycle breaking
        if break_combo_data is not None and len(break_combo_data["bond_break_combos"]):
            ring_bonds = []
            for ring in setup.rings:
                for bond in setup.get_bonds_in_ring(ring):
                    ring_bonds.append(bond)
            flex_model["torsions_org"] = 0
            for bond in setup.bond_info:
                if setup.bond_info[bond].rotatable and bond not in ring_bonds:
                    flex_model["torsions_org"] += 1

        setup.flexibility_model = flex_model

        # add G pseudo atoms and set CG types
        update_closure_atoms(setup, bonds_to_break, glue_pseudo_atoms)

        return

    @staticmethod
    def get_atom_params(
        input_atom_params, load_atom_params, add_atom_types, packaged_params
    ):
        """

        Parameters
        ----------
        input_atom_params
        load_atom_params
        add_atom_types
        packaged_params

        Returns
        -------

        """
        atom_params = {}
        if type(load_atom_params) == str:
            load_atom_params = [load_atom_params]
        elif load_atom_params is None:
            load_atom_params = ()
        for name in load_atom_params:
            filename = None
            if (
                name == "openff-2.0.0" or name == "openff"
            ):  # TODO allow multiple versions
                vdw_list, _, _ = load_openff()
                d = {"openff-2.0.0": vdw_list}
            elif name in packaged_params:
                filename = packaged_params[name]
            elif name.endswith(".json"):
                filename = name
            else:
                msg = (
                    "names passed to 'load_atom_params' need to suffixed with .json"
                    + eol
                )
                msg += (
                    "or be the unsuffixed basename of a JSON file in %s."
                    % str(params_dir)
                    + eol
                )
                msg += "name was %s" % name
                raise ValueError(msg)
            if filename is not None:
                with open(filename) as f:
                    d = json.load(f)
            overlapping_groups = set(atom_params).intersection(set(d))
            if len(overlapping_groups):
                msg = "overlapping parameter groups: %s" % str(overlapping_groups)
                raise ValueError(msg)
            atom_params.update(d)

        if input_atom_params is not None:
            d = json.loads(json.dumps(input_atom_params))
            overlapping_groups = set(atom_params).intersection(set(d))
            if len(overlapping_groups):  # todo: remove duplicated code
                msg = "overlapping parameter groups: %s" % str(overlapping_groups)
                raise ValueError(msg)
            params_set_here = set()
            for group_key, rows in input_atom_params.items():
                for row in rows:
                    for param_name in row:
                        if param_name not in ["smarts", "comment", "IDX"]:
                            params_set_here.add(param_name)
            params_set_before = set()
            for group_key, rows in atom_params.items():
                for row in rows:
                    for param_name in row:
                        if param_name not in ["smarts", "comment", "IDX"]:
                            params_set_before.add(param_name)
            overlap = params_set_before.intersection(params_set_here)
            if len(overlap):
                msg = f"input_atom_params {overlap} also set by one or more of {load_atom_params}\n"
                msg += "consider setting load_atom_params=None"
                raise ValueError(
                    f"input_atom_params {overlap} also set by one or more of {load_atom_params}"
                )

            atom_params.update(d)

        if len(add_atom_types) > 0:
            group_keys = list(atom_params.keys())
            if len(group_keys) != 1:
                msg = "add_atom_types is usable only when there is one group of parameters"
                msg += ", but there are %d groups: %s" % (
                    len(group_keys),
                    str(group_keys),
                )
                raise RuntimeError(msg)
            key = group_keys[0]
            atom_params[key].extend(add_atom_types)

        return atom_params

    @property
    def setup(self):
        """

        Returns
        -------

        """
        msg = "MoleculePreparation.setup is deprecated in Meeko v0.5."
        msg += (
            " MoleculePreparation.prepare() returns a list of MoleculeSetup instances."
        )
        warnings.warn(msg, DeprecationWarning)
        if len(self.deprecated_setup_access) > 1:
            raise RuntimeError(
                "got multiple setups, use new api: molsetup_list = mk_prep(mol)"
            )

        return self.deprecated_setup_access[0]

    @classmethod
    def get_defaults_dict(cls):
        """

        Returns
        -------

        """
        defaults = {}
        sig = signature(cls)
        for key in sig.parameters:
            defaults[key] = sig.parameters[key].default
        return defaults

    def __call__(self, *args):
        return self.prepare(*args)

    def prepare(
        self,
        mol,
        root_atom_index=None,
        not_terminal_atoms=None,
        delete_ring_bonds=None,
        glue_pseudo_atoms=None,
        conformer_id=-1,
    ):
        """
        Create an RDKitMoleculeSetup from an RDKit Mol object.

        Parameters
        ----------
        mol: rdkit.Chem.rdchem.Mol
            An RDKit Mol with explicit hydrogens and 3D coordinates.
        root_atom_index: int
            Used to set ROOT of torsion tree instead of searching.
        not_terminal_atoms: list
            Makes bonds with terminal atoms rotatable (e.g. C-Alpha carbon in flexres).
        delete_ring_bonds: list[tuple[int, int]]
            Bonds deleted for macrocycle flexibility. Each bond is a tuple of two ints (atom 0-indices).
        glue_pseudo_atoms: dict
            Mapping from parent atom indices to coordinates.
        conformer_id: int

        Returns
        -------
        setups: list[RDKitMoleculeSetup]
            Returns a list of generated RDKitMoleculeSetups
        """

        if not_terminal_atoms is None:
            not_terminal_atoms = []
        if delete_ring_bonds is None:
            delete_ring_bonds = []
        if glue_pseudo_atoms is None:
            glue_pseudo_atoms = {}
        mol_type = type(mol)
        if mol_type not in self._classes_setup:
            raise TypeError(
                "Molecule is not an instance of supported types: %s" % type(mol)
            )
        setup_class = self._classes_setup[mol_type]
        setup = setup_class.from_mol(
            mol,
            keep_chorded_rings=self.keep_chorded_rings,
            keep_equivalent_rings=self.keep_equivalent_rings,
            assign_charges=self.charge_model == "gasteiger",
            conformer_id=conformer_id,
        )

        self.check_external_ring_break(setup, delete_ring_bonds, glue_pseudo_atoms)

        # 1.  assign atom params
        AtomTyper.type_everything(
            setup,
            self.atom_params,
            self.charge_model,
            self.offatom_params,
            self.dihedral_params,
        )

        # Convert molecule to graph and apply trained Espaloma model
        if self.dihedral_model == "espaloma" or self.charge_model == "espaloma":
            molgraph = self.espaloma_model.get_espaloma_graph(setup)

        # Grab dihedrals from graph node and set them to the molsetup
        if self.dihedral_model == "espaloma":
            self.espaloma_model.set_espaloma_dihedrals(setup, molgraph)

        # Grab charges from graph node and set them to the molsetup
        if self.charge_model == "espaloma":
            self.espaloma_model.set_espaloma_charges(setup, molgraph)

        # merge hydrogens (or any terminal atoms)
        indices = set()
        for atype_to_merge in self.merge_these_atom_types:
            for atom in setup.atoms:
                if atom.atom_type == atype_to_merge:
                    indices.add(atom.index)
        setup.merge_terminal_atoms(indices)

        # 3.  assign bond types
        #     - all single bonds rotatable except some amides and SMARTS rigidification
        #     - macrocycle code breaks rings only at rotatable bonds
        #     - bonds in rigid rings are set as non-rotatable after the flex_model is built
        self._bond_typer(
            setup,
            self.flexible_amides,
            self.rigidify_bonds_smarts,
            self.rigidify_bonds_indices,
        )

        # 4 . hydrate molecule
        if self.hydrate:
            self._water_builder.hydrate(setup)

        self.calc_flex(
            setup,
            root_atom_index,
            not_terminal_atoms,
            delete_ring_bonds,
            glue_pseudo_atoms,
        )

        if self.reactive_smarts is None:
            setups = [setup]
        else:
            reactive_types_dicts = assign_reactive_types(
                setup,
                self.reactive_smarts,
                self.reactive_smarts_idx,
            )

            if len(reactive_types_dicts) == 0:
                raise RuntimeError("reactive SMARTS didn't match")

            setups = []
            for r in reactive_types_dicts:
                new_setup = setup.copy()
                # There is no guarantee that the addition order in the dictionary will be the correct order to
                # create the list in, so first sorts the keys from the dictionary then extracts the values in order
                # to construct the new atom type list.
                for idx, atom_type in r.items():
                    new_setup.atoms[idx].atom_type = atom_type
                setups.append(new_setup)

        # for a gentle introduction of the new API
        self.deprecated_setup_access = setups

        return setups

    @staticmethod
    def check_external_ring_break(molsetup, break_ring_bonds, glue_pseudo_atoms):
        """

        Parameters
        ----------
        molsetup: RDKitMoleculeSetup
        break_ring_bonds:
        glue_pseudo_atoms: dict

        Returns
        -------
        None

        Raises
        ------
        ValueError:
            If bonds are missing from the MoleculeSetup, if glue_pseudo_atoms is missing certain atom indices, and if
            there is an incorrect number of coordinates in glue_pseudo_atoms.
        """
        for index1, index2 in break_ring_bonds:
            has_bond = Bond.get_bond_id(index1, index2) in molsetup.bond_info
            if not has_bond:
                raise ValueError("bond (%d, %d) not in molsetup" % (index1, index2))
            for index in (index1, index2):
                if index not in glue_pseudo_atoms:
                    raise ValueError("missing glue pseudo for atom %d" % index)
                xyz = glue_pseudo_atoms[index]
                if len(xyz) != 3:
                    raise ValueError(
                        "expected 3 coordinates (got %d) for glue pseudo of atom %d"
                        % (len(xyz), index)
                    )
        return

    def write_pdbqt_string(self):
        """
        Writes a PDBQT string. Deprecated in Meeko v0.5.

        Returns
        -------

        """
        msg = "MoleculePreparation.write_pdbqt_string() is deprecated in Meeko v0.5."
        msg += " Pass the MoleculeSetup instance to PDBQTWriterLegacy.write_string()."
        msg += (
            " MoleculePreparation.prepare() returns a list of MoleculeSetup instances."
        )
        warnings.warn(msg, DeprecationWarning)
        pdbqt_string, is_ok, err_msg = PDBQTWriterLegacy.write_string(self.setup)
        if not is_ok:
            msg = "Cannot generate PDBQT, error from PDBQTWriterLegacy:" + eol
            msg += err_msg
            raise RuntimeError(msg)
        return pdbqt_string

    def write_pdbqt_file(self, pdbqt_filename):
        """
        Writes out a pdbqt file. Deprecated in Meeko v0.5

        Parameters
        ----------
        pdbqt_filename: str
            PDBQT filename to write to

        Returns
        -------
        None
        """
        warnings.warn(
            "MoleculePreparation.write_pdbqt_file() is deprecated since Meeko v0.5",
            DeprecationWarning,
        )
        with open(pdbqt_filename, "w") as w:
            w.write(self.write_pdbqt_string())



from rdkit import Chem
from rdkit.Chem import rdChemReactions
from .utils import mini_periodic_table
from .pdbutils import PDBAtomInfo
from rdkit.Geometry import Point3D
from rdkit.Chem import rdDetermineBonds

periodic_table = Chem.GetPeriodicTable()


"""
create new RDKIT residue

mi  =  Chem.AtomPDBResidueInfo()
mi.SetResidueName('MOL')
mi.SetResidueNumber(1)
mi.SetOccupancy(0.0)
mi.SetTempFactor(0.0)

source: https://sourceforge.net/p/rdkit/mailman/message/36404394/
"""


def getPdbInfoNoNull(atom):
    """extract information for populating an ATOM/HETATM line
    in the PDB"""
    minfo = atom.GetMonomerInfo()  # same as GetPDBResidueInfo
    if minfo is None:
        atomic_number = atom.GetAtomicNum()
        if atomic_number == 0:
            name = "%-2s" % "*"
        else:
            name = "%-2s" % mini_periodic_table[atomic_number]
        chain = " "
        resNum = 1
        icode = ""
        resName = "UNL"
    else:
        name = minfo.GetName()
        chain = minfo.GetChainId()
        resNum = minfo.GetResidueNumber()
        icode = minfo.GetInsertionCode()
        resName = minfo.GetResidueName()
    return PDBAtomInfo(
        name=name, resName=resName, resNum=resNum, icode=icode, chain=chain
    )


class Mol2MolSupplier:
    """RDKit Mol2 molecule supplier.
    Parameters
        sanitize: perform RDKit sanitization of Mol2 molecule"""

    def __init__(
        self, filename, sanitize=True, removeHs=False, cleanupSubstructures=True
    ):
        self.fp = open(filename, "r")
        self._opts = {
            "sanitize": sanitize,
            "removeHs": removeHs,
            "cleanupSubstructures": cleanupSubstructures,
        }
        self.buff = []

    def __iter__(self):
        return self

    def __next__(self):
        """iterator step"""
        while True:
            line = self.fp.readline()
            # empty line
            if not line:
                if len(self.buff):
                    # buffer full, returning last molecule
                    mol = Chem.MolFromMol2Block("".join(self.buff), **self._opts)
                    self.buff = []
                    return mol
                # buffer empty, stopping the iteration
                self.fp.close()
                raise StopIteration
            if "@<TRIPOS>MOLECULE" in line:
                # first molecule parsed
                if len(self.buff) == 0:
                    self.buff.append(line)
                else:
                    # found the next molecule, breaking to return the complete one
                    break
            else:
                # adding another line in the current molecule
                self.buff.append(line)
        # found a complete molecule, returning it
        mol = Chem.MolFromMol2Block("".join(self.buff), **self._opts)
        self.buff = [line]
        return mol

class AtomField:
    """Stores data parsed from PDB or mmCIF"""

    def __init__(
        self,
        atomname: str,
        altloc: str,
        resname: str,
        chain: str,
        resnum: int,
        icode: str,
        x: float,
        y: float,
        z: float,
        element: str,
    ):
        self.atomname = atomname
        self.altloc = altloc
        self.resname = resname
        self.chain = chain
        self.resnum = resnum
        self.icode = icode
        self.x = x
        self.y = y
        self.z = z
        if len(element) > 1:
            element = f"{element[0].upper()}{element[1].lower()}"
        else:
            element = f"{element.upper()}"
        self.atomic_nr = periodic_table.GetAtomicNumber(element)


def _build_rdkit_mol_for_altloc(atom_fields_list, wanted_altloc:str=None):
    mol = Chem.EditableMol(Chem.Mol())
    mol.BeginBatchEdit() 
    positions = []
    idx_to_rdkit = {}
    for index_list, atom in enumerate(atom_fields_list):
        if wanted_altloc is not None:
            if atom.altloc and atom.altloc != wanted_altloc:
                # if atom.altloc is "" we still want to consider this atom
                continue
        rdkit_atom = Chem.Atom(atom.atomic_nr)
        positions.append(Point3D(atom.x, atom.y, atom.z))
        res_info = Chem.AtomPDBResidueInfo()
        res_info.SetName(atom.atomname)
        res_info.SetResidueName(atom.resname)
        res_info.SetResidueNumber(atom.resnum)
        res_info.SetChainId(atom.chain)
        res_info.SetInsertionCode(atom.icode)
        rdkit_atom.SetPDBResidueInfo(res_info)
        index_rdkit = mol.AddAtom(rdkit_atom)
        idx_to_rdkit[index_list] = index_rdkit
    mol.CommitBatchEdit()
    mol = mol.GetMol()
    conformer = Chem.Conformer(mol.GetNumAtoms())
    for index, position in enumerate(positions):
        conformer.SetAtomPosition(index, position)
    mol.AddConformer(conformer, assignId=True)
    return mol, idx_to_rdkit
        

def build_one_rdkit_mol_per_altloc(atom_fields_list):
    """ if no altlocs, the only key in the output dict is None
        if altlocs exist, None is not a key: the keys are the altloc IDs
    """
    altlocs = set([atom.altloc for atom in atom_fields_list if atom.altloc])
    rdkit_mol_dict = {}
    if not altlocs:
        altlocs = {None}
    for altloc in altlocs:
        mol, idx_to_rdkit = _build_rdkit_mol_for_altloc(atom_fields_list, altloc)
        rdkit_mol_dict[altloc] = (mol, idx_to_rdkit)
    return rdkit_mol_dict


def _aux_altloc_mol_build(atom_field_list, requested_altloc, default_altloc):
    missed_altloc = False
    needed_altloc = False
    mols_dict = build_one_rdkit_mol_per_altloc(atom_field_list) 
    has_altloc = None not in mols_dict
    if has_altloc and requested_altloc is None and default_altloc is None:
        pdbmol = None
        missed_altloc = False 
        needed_altloc = True
    elif requested_altloc and requested_altloc in mols_dict:
        pdbmol, idx_to_rdkit = mols_dict[requested_altloc]
    elif requested_altloc and requested_altloc not in mols_dict:
        pdbmol = None
        missed_altloc = True
        needed_altloc = False
    elif default_altloc and default_altloc in mols_dict:
        pdbmol, idx_to_rdkit = mols_dict[default_altloc]
    elif has_altloc and default_altloc not in mols_dict:
        pdbmol = None
        missed_altloc = True
        needed_altloc = False
    elif not has_altloc and requested_altloc is None:
        pdbmol, idx_to_rdkit = mols_dict[None]
    else:
        raise RuntimeError("programming bug, please post full error on github")
    if pdbmol is None: 
        idx_to_rdkit = None
        return pdbmol, idx_to_rdkit, missed_altloc, needed_altloc
    else:
        rdDetermineBonds.DetermineConnectivity(pdbmol)
        for atom in pdbmol.GetAtoms():
            if atom.GetAtomicNum() == 7 and len(atom.GetNeighbors()) == 4:
                atom.SetFormalCharge(1)
        _ = Chem.SanitizeMol(pdbmol)

    return pdbmol, idx_to_rdkit, missed_altloc, needed_altloc

def react_and_map(reactants: tuple[Chem.Mol], rxn: rdChemReactions.ChemicalReaction):
    """
    Run a reaction and keep track of atom indices from reactants to products.
    
    Parameters
    ----------
    reactants : tuple[Chem.Mol]
        A tuple of RDKit molecule objects representing the reactants.
    rxn : rdChemReactions.ChemicalReaction
        The RDKit reaction object.
        
    Returns
    -------
    list[tuple[Chem.Mol, dict[str, list[Optional[int]]]]]
        A list of tuples where each tuple contains a product molecule and a dictionary.
        The dictionary has keys 'atom_idx' and 'new_atom_label', which are ordered lists for product atoms:
        - 'atom_idx' holds the corresponding atom indices in reactant. None for newly added atoms. 
        - 'new_atom_label' holds the reaction mapping number, only for newly added atoms. 
    """

    # Prepare for multiple possible outcomes resulted from multiple matched reactive sites in reactant
    outcomes = []
    for products in rxn.RunReactants(reactants): 
        # Assumes single product 
        product = products[0]
        # For each atom, get react_atom_idx if they were in reactant
        atom_idxmap = [
            atom.GetIntProp("react_atom_idx") if atom.HasProp("react_atom_idx")
            else None
            for atom in product.GetAtoms()
        ]
        # For each atom, get the rxn mapping number if the were added in the rxn
        new_atom_label = [
            atom.GetIntProp("old_mapno") if atom.HasProp("old_mapno") and not atom.HasProp("react_atom_idx")
            else None
            for atom in product.GetAtoms()
        ]
        # Collect product and index_map
        index_map = {"atom_idx": atom_idxmap, "new_atom_label": new_atom_label}
        outcomes.append((product, index_map))

    return outcomes


covalent_radius = {  # from wikipedia
    1: 0.31,
    5: 0.84,
    6: 0.76,
    7: 0.71,
    8: 0.66,
    9: 0.57,
    12: 0.00,  # hack to avoid bonds with metals
    14: 1.11,
    15: 1.07,
    16: 1.05,
    17: 1.02,
    # 19: 2.03,
    20: 0.00,
    # 24: 1.39,
    25: 0.00,  # hack to avoid bonds with metals
    26: 0.00,
    30: 0.00,  # hack to avoid bonds with metals
    # 34: 1.20,
    35: 1.20,
    53: 1.39,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko bond typer
#


class BondTyperLegacy:

    def __call__(
        self, setup, flexible_amides, rigidify_bonds_smarts, rigidify_bonds_indices
    ):
        """Typing atom bonds in the legacy way

        Args:
            setup: MoleculeSetup object

            rigidify_bond_smarts (list): patterns to freeze bonds, e.g. conjugated carbons
        """

        amide_bonds = [
            (x[0], x[1]) for x in setup.find_pattern("[NX3]-[CX3]=[O,N,S]")
        ]  # includes amidines

        # tertiary amides with non-identical substituents will be allowed to rotate
        tertiary_amides = [
            x for x in setup.find_pattern("[NX3]([!#1])([!#1])-[CX3]=[O,N,S]")
        ]
        equivalent_atoms = setup.get_equivalent_atoms()
        num_amides_removed = 0
        num_amides_originally = len(amide_bonds)
        for x in tertiary_amides:
            r1, r2 = x[1], x[2]
            if equivalent_atoms[r1] != equivalent_atoms[r2]:
                amide_bonds.remove((x[0], x[3]))
                num_amides_removed += 1
        assert num_amides_originally == num_amides_removed + len(amide_bonds)

        to_rigidify = set()
        n_smarts = len(rigidify_bonds_smarts)
        assert n_smarts == len(rigidify_bonds_indices)
        for i in range(n_smarts):
            a, b = rigidify_bonds_indices[i]
            smarts = rigidify_bonds_smarts[i]
            indices_list = setup.find_pattern(smarts)
            for indices in indices_list:
                atom_a = indices[a]
                atom_b = indices[b]
                to_rigidify.add((atom_a, atom_b))
                to_rigidify.add((atom_b, atom_a))

        for bond_id, bond in setup.bond_info.items():
            rotatable = True
            if bond_id in to_rigidify:
                rotatable = False
            # check if bond is amide
            if (
                bond_id in amide_bonds
                or (bond_id[1], bond_id[0]) in amide_bonds
            ) and not flexible_amides:
                rotatable = False
            bond.rotatable = rotatable



from meeko import RDKitMolCreate
from meeko import PDBQTMolecule
from meeko import MoleculePreparation
from meeko import PDBQTWriterLegacy
from rdkit import Chem
import numpy as np
import pathlib
import json
import io

workdir = pathlib.Path(__file__)
datadir = workdir.parents[0] / "rdkitmol_from_docking_data"

# test for reasonable bond lengths. This will fail if the atom names
# in RDKiMolCreate.flexres[<resname>]["atom_names"] are in the wrong
# order w.r.t RDKitMolCreate.flexres[<resname>]["smiles"]
bond_range_H = [0.85, 1.15]
bond_range_no_H = [1.16, 1.65]
bond_S_bonus = 0.2

def check_rdkit_bond_lengths(fpath, nr_expected_none, is_dlg, skip_typing):
    pdbqtmol = PDBQTMolecule.from_file(fpath, is_dlg=is_dlg, skip_typing=skip_typing)
    return run_from_pdbqtmol(pdbqtmol, nr_expected_none)

def run_from_pdbqtmol(pdbqtmol, nr_expected_none=0):
    mols = RDKitMolCreate.from_pdbqt_mol(pdbqtmol, keep_flexres=True)
    assert(mols.count(None) == nr_expected_none)
    nr_conformers = set()
    for mol in mols:
        if mol is None:
            continue
        nr_conformers.add(mol.GetNumConformers())
        assert(len(nr_conformers) == 1)
        for confid in range(list(nr_conformers)[0]):
            positions = mol.GetConformer(confid).GetPositions()
            for bond in mol.GetBonds():
                has_H = bond.GetBeginAtom().GetAtomicNum() == 1
                has_H = has_H or (bond.GetEndAtom().GetAtomicNum() == 1)
                bond_range = bond_range_H if has_H else bond_range_no_H
                has_S = bond.GetBeginAtom().GetAtomicNum() in [16, 17, 35]
                has_S = has_S or (bond.GetEndAtom().GetAtomicNum() in [16, 17, 35])
                if has_S:
                    bond_range = [value + bond_S_bonus for value in bond_range] 
                a = positions[bond.GetBeginAtomIdx(), :]
                b = positions[bond.GetEndAtomIdx(), :]
                dist = sum([(a[i]-b[i])**2 for i in range(3)])**0.5
                assert(dist > bond_range[0])
                assert(dist < bond_range[1])

def test_asn_phe():
    fpath = datadir / "macrocycle-water-asn-phe.pdbqt"
    check_rdkit_bond_lengths(fpath, nr_expected_none=0, is_dlg=False, skip_typing=True)

def test_22_flexres():
    fpath = datadir / "22-flexres.pdbqt"
    check_rdkit_bond_lengths(fpath, nr_expected_none=0, is_dlg=False, skip_typing=True)

def test_phe_badphe():
    fpath = datadir / "arg_gln_asn_phe_badphe.pdbqt"
    check_rdkit_bond_lengths(fpath, nr_expected_none=1, is_dlg=False, skip_typing=True)

def test_arg_his():
    fpath = datadir / "arg_his.pdbqt"
    check_rdkit_bond_lengths(fpath, nr_expected_none=0, is_dlg=False, skip_typing=True)


# The following tests  generate the PDBQT and convert it back to RDKit,
# as opposed to the tests above which start from PDBQT.

mk_prep = MoleculePreparation()
mk_prep_wet = MoleculePreparation(hydrate=True)

def run(sdfname, wet=False):
    fpath = datadir / sdfname
    for mol in Chem.SDMolSupplier(str(fpath), removeHs=False):
        if wet:
            setups = mk_prep_wet.prepare(mol)
            pdbqt, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
        else:
            setups = mk_prep.prepare(mol)
            pdbqt, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
        pmol = PDBQTMolecule(pdbqt)
        run_from_pdbqtmol(pmol)

def test_small_01_zero_deuterium(): run("small-01_zero-deuterium.sdf")
def test_small_01_one_deuterium(): run("small-01_one-deuterium.sdf")
def test_small_01_two_deuterium(): run("small-01_two-deuterium.sdf")
def test_small_01_three_deuterium(): run("small-01_three-deuterium.sdf")
def test_small_01_four_deuterium(): run("small-01_four-deuterium.sdf")

def test_small_02_zero_deuterium(): run("small-02_zero-deuterium.sdf")
def test_small_02_one_deuterium_A(): run("small-02_one-deuterium-A.sdf")
def test_small_02_one_deuterium_B(): run("small-02_one-deuterium-B.sdf")
def test_small_02_one_deuterium_C(): run("small-02_one-deuterium-C.sdf")
def test_small_02_two_deuterium_A(): run("small-02_two-deuterium-A.sdf")
def test_small_02_two_deuterium_B(): run("small-02_two-deuterium-B.sdf")
def test_small_02_two_deuterium_C(): run("small-02_two-deuterium-C.sdf", wet=True)
def test_small_02_two_deuterium_D(): run("small-02_two-deuterium-D.sdf", wet=True)
def test_small_02_five_deuterium(): run("small-02_five-deuterium.sdf", wet=True)

def test_small_03_zero_deuterium(): run("small-03_zero-deuterium.sdf", wet=True)
def test_small_03_one_deuterium_A(): run("small-03_one-deuterium-A.sdf", wet=True)
def test_small_03_one_deuterium_B(): run("small-03_one-deuterium-B.sdf", wet=True)
def test_small_03_two_deuterium(): run("small-03_two-deuterium.sdf")
def test_small_03_three_deuterium(): run("small-03_three-deuterium.sdf")

def test_small_04(): run("small-04.sdf", wet=True)

def test_meeko_free_energy_prop_vina():
    fpath = datadir/ "vina-result-ethanol.pdbqt"
    pdbqt_mol = PDBQTMolecule.from_file(fpath)
    ref_charges = [0.034, 0.152, -0.397, 0.21]
    tolerance = 0.01
    assert np.all(np.abs(pdbqt_mol.atoms()["partial_charges"] - ref_charges) < tolerance)
    sd_string, failures = RDKitMolCreate.write_sd_string(pdbqt_mol)
    assert len(failures) == 0
    sio = io.BytesIO(sd_string.encode())
    nr_mols = 0
    for rdkit_mol in Chem.ForwardSDMolSupplier(sio):
        nr_mols += 1
    assert nr_mols == 1
    data = json.loads(rdkit_mol.GetProp("meeko"))
    assert "free_energy" in data
    assert data["free_energy"] == -2.7

def test_meeko_free_energy_prop_adgpu():
    fpath = datadir/ "adgpu-result-ethanol.dlg"
    pdbqt_mol = PDBQTMolecule.from_file(fpath, is_dlg=True)
    sd_string, failures = RDKitMolCreate.write_sd_string(pdbqt_mol, only_cluster_leads=True)
    assert len(failures) == 0
    sio = io.BytesIO(sd_string.encode())
    rdkit_mols = [mol for mol in Chem.ForwardSDMolSupplier(sio)]
    assert len(rdkit_mols) == 4 # 20 runs, 4 clusters
    data = json.loads(rdkit_mols[0].GetProp("meeko"))
    assert "free_energy" in data
    assert data["free_energy"] == -2.11

def test_offsite_charges():
    mk_config = {
        "input_offatom_params": {
            "offsite": [
                {"smarts": "[#7X3;v3;!+]([*])([*])[*]",
                 "IDX": [0],
                 "OFFATOMS": [{"z": [1, 2, 3], "phi": 0, "distance": 0.2, "atype": "LP", "pull_charge_fraction": 1.08}]
                 }
            ]
        }
    }

    sdfname = "small-01_three-deuterium.sdf"
    fpath = datadir / sdfname
    mk_prep = MoleculePreparation.from_config(mk_config)
    for mol in Chem.SDMolSupplier(str(fpath), removeHs=False):
        break # we have just 1 mol
    setups = mk_prep.prepare(mol)
    molsetup = setups[0]
    pdbqt, is_ok, error_msg = PDBQTWriterLegacy.write_string(molsetup)
    assert is_ok
    assert "LP" in pdbqt
    pmol = PDBQTMolecule(pdbqt, skip_typing=True)
    rdkit_mols = RDKitMolCreate.from_pdbqt_mol(pmol, keep_flexres=True)
    assert len(rdkit_mols) == 1
    smiles = Chem.MolToSmiles(rdkit_mols[0])
    assert type(smiles) == str
    assert len(smiles) > 0
